from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable, Iterator, Literal, Optional, Union

import pandas as pd
from pandas.io.formats.style import Styler

from pixeltable import DataFrame, catalog, exceptions as excs, exprs, func, share, type_system as ts
from pixeltable.catalog import Catalog, TableVersionPath
from pixeltable.catalog.insertable_table import OnErrorParameter
from pixeltable.config import Config
from pixeltable.env import Env
from pixeltable.iterators import ComponentIterator

if TYPE_CHECKING:
    import datasets  # type: ignore[import-untyped]

    RowData = list[dict[str, Any]]
    TableDataSource = Union[
        str,
        os.PathLike,
        Path,  # OS paths, filenames, URLs
        Iterator[dict[str, Any]],  # iterator producing dictionaries of values
        RowData,  # list of dictionaries
        DataFrame,  # Pixeltable DataFrame
        pd.DataFrame,  # pandas DataFrame
        'datasets.Dataset',
        'datasets.DatasetDict',  # Huggingface datasets
    ]


_logger = logging.getLogger('pixeltable')


def init(config_overrides: Optional[dict[str, Any]] = None) -> None:
    """Initializes the Pixeltable environment."""
    if config_overrides is None:
        config_overrides = {}
    Config.init(config_overrides)
    _ = Catalog.get()


def create_table(
    path: str,
    schema: Optional[dict[str, Any]] = None,
    *,
    source: Optional[TableDataSource] = None,
    source_format: Optional[Literal['csv', 'excel', 'parquet', 'json']] = None,
    schema_overrides: Optional[dict[str, Any]] = None,
    on_error: Literal['abort', 'ignore'] = 'abort',
    primary_key: Optional[Union[str, list[str]]] = None,
    num_retained_versions: int = 10,
    comment: str = '',
    media_validation: Literal['on_read', 'on_write'] = 'on_write',
    if_exists: Literal['error', 'ignore', 'replace', 'replace_force'] = 'error',
    extra_args: Optional[dict[str, Any]] = None,  # Additional arguments to data source provider
) -> catalog.Table:
    """Create a new base table. Exactly one of `schema` or `source` must be provided.

    If a `schema` is provided, then an empty table will be created with the specified schema.

    If a `source` is provided, then Pixeltable will attempt to infer a data source format and table schema from the
    contents of the specified data, and the data will be imported from the specified source into the new table. The
    source format and/or schema can be specified directly via the `source_format` and `schema_overrides` parameters.

    Args:
        path: Pixeltable path (qualified name) of the table, such as `'my_table'` or `'my_dir.my_subdir.my_table'`.
        schema: Schema for the new table, mapping column names to Pixeltable types.
        source: A data source (file, URL, DataFrame, or list of rows) to import from.
        source_format: Must be used in conjunction with a `source`.
            If specified, then the given format will be used to read the source data. (Otherwise,
            Pixeltable will attempt to infer the format from the source data.)
        schema_overrides: Must be used in conjunction with a `source`.
            If specified, then columns in `schema_overrides` will be given the specified types.
            (Pixeltable will attempt to infer the types of any columns not specified.)
        on_error: Determines the behavior if an error occurs while evaluating a computed column or detecting an
            invalid media file (such as a corrupt image) for one of the inserted rows.

            - If `on_error='abort'`, then an exception will be raised and the rows will not be inserted.
            - If `on_error='ignore'`, then execution will continue and the rows will be inserted. Any cells
                with errors will have a `None` value for that cell, with information about the error stored in the
                corresponding `tbl.col_name.errortype` and `tbl.col_name.errormsg` fields.
        primary_key: An optional column name or list of column names to use as the primary key(s) of the
            table.
        num_retained_versions: Number of versions of the table to retain.
        comment: An optional comment; its meaning is user-defined.
        media_validation: Media validation policy for the table.

            - `'on_read'`: validate media files at query time
            - `'on_write'`: validate media files during insert/update operations
        if_exists: Determines the behavior if a table already exists at the specified path location.

            - `'error'`: raise an error
            - `'ignore'`: do nothing and return the existing table handle
            - `'replace'`: if the existing table has no views or snapshots, drop and replace it with a new one;
                raise an error if the existing table has views or snapshots
            - `'replace_force'`: drop the existing table and all its views and snapshots, and create a new one
        extra_args: Must be used in conjunction with a `source`. If specified, then additional arguments will be
            passed along to the source data provider.

    Returns:
        A handle to the newly created table, or to an already existing table at the path when `if_exists='ignore'`.
            Please note the schema of the existing table may not match the schema provided in the call.

    Raises:
        Error: if

            - the path is invalid, or
            - the path already exists and `if_exists='error'`, or
            - the path already exists and is not a table, or
            - an error occurs while attempting to create the table, or
            - an error occurs while attempting to import data from the source.

    Examples:
        Create a table with an int and a string column:

        >>> tbl = pxt.create_table('my_table', schema={'col1': pxt.Int, 'col2': pxt.String})

        Create a table from a select statement over an existing table `orig_table` (this will create a new table
        containing the exact contents of the query):

        >>> tbl1 = pxt.get_table('orig_table')
        ... tbl2 = pxt.create_table('new_table', tbl1.where(tbl1.col1 < 10).select(tbl1.col2))

        Create a table if it does not already exist, otherwise get the existing table:

        >>> tbl = pxt.create_table('my_table', schema={'col1': pxt.Int, 'col2': pxt.String}, if_exists='ignore')

        Create a table with an int and a float column, and replace any existing table:

        >>> tbl = pxt.create_table('my_table', schema={'col1': pxt.Int, 'col2': pxt.Float}, if_exists='replace')

        Create a table from a CSV file:

        >>> tbl = pxt.create_table('my_table', source='data.csv')
    """
    from pixeltable.io.table_data_conduit import DFTableDataConduit, UnkTableDataConduit
    from pixeltable.io.utils import normalize_primary_key_parameter

    if (schema is None) == (source is None):
        raise excs.Error('Either a `schema` or a `source` must be provided (but not both)')

    if schema is not None and (len(schema) == 0 or not isinstance(schema, dict)):
        raise excs.Error('`schema` must be a non-empty dictionary')

    path_obj = catalog.Path(path)
    if_exists_ = catalog.IfExistsParam.validated(if_exists, 'if_exists')
    media_validation_ = catalog.MediaValidation.validated(media_validation, 'media_validation')
    primary_key: Optional[list[str]] = normalize_primary_key_parameter(primary_key)
    table: catalog.Table = None
    tds = None
    data_source = None
    if source is not None:
        tds = UnkTableDataConduit(source, source_format=source_format, extra_fields=extra_args)
        tds.check_source_format()
        data_source = tds.specialize()
        src_schema_overrides: dict[str, ts.ColumnType] = {}
        if schema_overrides is not None:
            for col_name, py_type in schema_overrides.items():
                col_type = ts.ColumnType.normalize_type(py_type, nullable_default=True, allow_builtin_types=False)
                if col_type is None:
                    raise excs.Error(f'Invalid type for column {col_name!r} in `schema_overrides`: {py_type}')
                src_schema_overrides[col_name] = col_type
        data_source.src_schema_overrides = src_schema_overrides
        data_source.src_pk = primary_key
        data_source.infer_schema()
        schema = data_source.pxt_schema
        primary_key = data_source.pxt_pk
        is_direct_df = data_source.is_direct_df()
    else:
        is_direct_df = False

    if len(schema) == 0 or not isinstance(schema, dict):
        raise excs.Error(
            'Unable to create a proper schema from supplied `source`. Please use appropriate `schema_overrides`.'
        )

    table = Catalog.get().create_table(
        path_obj,
        schema,
        data_source.pxt_df if isinstance(data_source, DFTableDataConduit) else None,
        if_exists=if_exists_,
        primary_key=primary_key,
        comment=comment,
        media_validation=media_validation_,
        num_retained_versions=num_retained_versions,
    )
    if data_source is not None and not is_direct_df:
        fail_on_exception = OnErrorParameter.fail_on_exception(on_error)
        table.insert_table_data_source(data_source=data_source, fail_on_exception=fail_on_exception)

    return table


def create_view(
    path: str,
    base: Union[catalog.Table, DataFrame],
    *,
    additional_columns: Optional[dict[str, Any]] = None,
    is_snapshot: bool = False,
    iterator: Optional[tuple[type[ComponentIterator], dict[str, Any]]] = None,
    num_retained_versions: int = 10,
    comment: str = '',
    media_validation: Literal['on_read', 'on_write'] = 'on_write',
    if_exists: Literal['error', 'ignore', 'replace', 'replace_force'] = 'error',
) -> Optional[catalog.Table]:
    """Create a view of an existing table object (which itself can be a view or a snapshot or a base table).

    Args:
        path: A name for the view; can be either a simple name such as `my_view`, or a pathname such as
            `dir1.my_view`.
        base: [`Table`][pixeltable.Table] (i.e., table or view or snapshot) or [`DataFrame`][pixeltable.DataFrame] to
            base the view on.
        additional_columns: If specified, will add these columns to the view once it is created. The format
            of the `additional_columns` parameter is identical to the format of the `schema_or_df` parameter in
            [`create_table`][pixeltable.create_table].
        is_snapshot: Whether the view is a snapshot. Setting this to `True` is equivalent to calling
            [`create_snapshot`][pixeltable.create_snapshot].
        iterator: The iterator to use for this view. If specified, then this view will be a one-to-many view of
            the base table.
        num_retained_versions: Number of versions of the view to retain.
        comment: Optional comment for the view.
        media_validation: Media validation policy for the view.

            - `'on_read'`: validate media files at query time
            - `'on_write'`: validate media files during insert/update operations
        if_exists: Directive regarding how to handle if the path already exists.
            Must be one of the following:

            - `'error'`: raise an error
            - `'ignore'`: do nothing and return the existing view handle
            - `'replace'`: if the existing view has no dependents, drop and replace it with a new one
            - `'replace_force'`: drop the existing view and all its dependents, and create a new one

    Returns:
        A handle to the [`Table`][pixeltable.Table] representing the newly created view. If the path already
            exists and `if_exists='ignore'`, returns a handle to the existing view. Please note the schema
            or the base of the existing view may not match those provided in the call.

    Raises:
        Error: if

            - the path is invalid, or
            - the path already exists and `if_exists='error'`, or
            - the path already exists and is not a view, or
            - an error occurs while attempting to create the view.

    Examples:
        Create a view `my_view` of an existing table `my_table`, filtering on rows where `col1` is greater than 10:

        >>> tbl = pxt.get_table('my_table')
        ... view = pxt.create_view('my_view', tbl.where(tbl.col1 > 10))

        Create a view `my_view` of an existing table `my_table`, filtering on rows where `col1` is greater than 10,
        and if it not already exist. Otherwise, get the existing view named `my_view`:

        >>> tbl = pxt.get_table('my_table')
        ... view = pxt.create_view('my_view', tbl.where(tbl.col1 > 10), if_exists='ignore')

        Create a view `my_view` of an existing table `my_table`, filtering on rows where `col1` is greater than 100,
        and replace any existing view named `my_view`:

        >>> tbl = pxt.get_table('my_table')
        ... view = pxt.create_view('my_view', tbl.where(tbl.col1 > 100), if_exists='replace_force')
    """
    tbl_version_path: TableVersionPath
    select_list: Optional[list[tuple[exprs.Expr, Optional[str]]]] = None
    where: Optional[exprs.Expr] = None
    if isinstance(base, catalog.Table):
        tbl_version_path = base._tbl_version_path
        sample_clause = None
    elif isinstance(base, DataFrame):
        base._validate_mutable_op_sequence('create_view', allow_select=True)
        tbl_version_path = base._from_clause.tbls[0]
        where = base.where_clause
        sample_clause = base.sample_clause
        select_list = base.select_list
        if sample_clause is not None and not is_snapshot and not sample_clause.is_repeatable:
            raise excs.Error('Non-snapshot views cannot be created with non-fractional or stratified sampling')
    else:
        raise excs.Error('`base` must be an instance of `Table` or `DataFrame`')
    assert isinstance(base, (catalog.Table, DataFrame))

    path_obj = catalog.Path(path)
    if_exists_ = catalog.IfExistsParam.validated(if_exists, 'if_exists')
    media_validation_ = catalog.MediaValidation.validated(media_validation, 'media_validation')

    if additional_columns is None:
        additional_columns = {}
    else:
        # additional columns should not be in the base table
        for col_name in additional_columns:
            if col_name in [c.name for c in tbl_version_path.columns()]:
                raise excs.Error(
                    f'Column {col_name!r} already exists in the base table '
                    f'{tbl_version_path.get_column(col_name).tbl.name}.'
                )

    return Catalog.get().create_view(
        path_obj,
        tbl_version_path,
        select_list=select_list,
        where=where,
        sample_clause=sample_clause,
        additional_columns=additional_columns,
        is_snapshot=is_snapshot,
        iterator=iterator,
        num_retained_versions=num_retained_versions,
        comment=comment,
        media_validation=media_validation_,
        if_exists=if_exists_,
    )


def create_snapshot(
    path_str: str,
    base: Union[catalog.Table, DataFrame],
    *,
    additional_columns: Optional[dict[str, Any]] = None,
    iterator: Optional[tuple[type[ComponentIterator], dict[str, Any]]] = None,
    num_retained_versions: int = 10,
    comment: str = '',
    media_validation: Literal['on_read', 'on_write'] = 'on_write',
    if_exists: Literal['error', 'ignore', 'replace', 'replace_force'] = 'error',
) -> Optional[catalog.Table]:
    """Create a snapshot of an existing table object (which itself can be a view or a snapshot or a base table).

    Args:
        path_str: A name for the snapshot; can be either a simple name such as `my_snapshot`, or a pathname such as
            `dir1.my_snapshot`.
        base: [`Table`][pixeltable.Table] (i.e., table or view or snapshot) or [`DataFrame`][pixeltable.DataFrame] to
            base the snapshot on.
        additional_columns: If specified, will add these columns to the snapshot once it is created. The format
            of the `additional_columns` parameter is identical to the format of the `schema_or_df` parameter in
            [`create_table`][pixeltable.create_table].
        iterator: The iterator to use for this snapshot. If specified, then this snapshot will be a one-to-many view of
            the base table.
        num_retained_versions: Number of versions of the view to retain.
        comment: Optional comment for the snapshot.
        media_validation: Media validation policy for the snapshot.

            - `'on_read'`: validate media files at query time
            - `'on_write'`: validate media files during insert/update operations
        if_exists: Directive regarding how to handle if the path already exists.
            Must be one of the following:

            - `'error'`: raise an error
            - `'ignore'`: do nothing and return the existing snapshot handle
            - `'replace'`: if the existing snapshot has no dependents, drop and replace it with a new one
            - `'replace_force'`: drop the existing snapshot and all its dependents, and create a new one

    Returns:
        A handle to the [`Table`][pixeltable.Table] representing the newly created snapshot.
            Please note the schema or base of the existing snapshot may not match those provided in the call.

    Raises:
        Error: if

            - the path is invalid, or
            - the path already exists and `if_exists='error'`, or
            - the path already exists and is not a snapshot, or
            - an error occurs while attempting to create the snapshot.

    Examples:
        Create a snapshot `my_snapshot` of a table `my_table`:

        >>> tbl = pxt.get_table('my_table')
        ... snapshot = pxt.create_snapshot('my_snapshot', tbl)

        Create a snapshot `my_snapshot` of a view `my_view` with additional int column `col3`,
        if `my_snapshot` does not already exist:

        >>> view = pxt.get_table('my_view')
        ... snapshot = pxt.create_snapshot(
        ...     'my_snapshot', view, additional_columns={'col3': pxt.Int}, if_exists='ignore'
        ... )

        Create a snapshot `my_snapshot` on a table `my_table`, and replace any existing snapshot named `my_snapshot`:

        >>> tbl = pxt.get_table('my_table')
        ... snapshot = pxt.create_snapshot('my_snapshot', tbl, if_exists='replace_force')
    """
    return create_view(
        path_str,
        base,
        additional_columns=additional_columns,
        iterator=iterator,
        is_snapshot=True,
        num_retained_versions=num_retained_versions,
        comment=comment,
        media_validation=media_validation,
        if_exists=if_exists,
    )


def create_replica(destination: str, source: Union[str, catalog.Table]) -> Optional[catalog.Table]:
    """
    Create a replica of a table. Can be used either to create a remote replica of a local table, or to create a local
    replica of a remote table. A given table can have at most one replica per Pixeltable instance.

    Args:
        destination: Path where the replica will be created. Can be either a local path such as `'my_dir.my_table'`, or
            a remote URI such as `'pxt://username/mydir.my_table'`.
        source: Path to the source table, or (if the source table is a local table) a handle to the source table.
    """
    remote_dest = destination.startswith('pxt://')
    remote_source = isinstance(source, str) and source.startswith('pxt://')
    if remote_dest == remote_source:
        raise excs.Error('Exactly one of `destination` or `source` must be a remote URI.')

    if remote_dest:
        if isinstance(source, str):
            source = get_table(source)
        share.push_replica(destination, source)
        return None
    else:
        assert isinstance(source, str)
        return share.pull_replica(destination, source)


def get_table(path: str) -> catalog.Table:
    """Get a handle to an existing table, view, or snapshot.

    Args:
        path: Path to the table.

    Returns:
        A handle to the [`Table`][pixeltable.Table].

    Raises:
        Error: If the path does not exist or does not designate a table object.

    Examples:
        Get handle for a table in the top-level directory:

        >>> tbl = pxt.get_table('my_table')

        For a table in a subdirectory:

        >>> tbl = pxt.get_table('subdir.my_table')

        Handles to views and snapshots are retrieved in the same way:

        >>> tbl = pxt.get_table('my_snapshot')
    """
    path_obj = catalog.Path(path)
    tbl = Catalog.get().get_table(path_obj)
    return tbl


def move(path: str, new_path: str) -> None:
    """Move a schema object to a new directory and/or rename a schema object.

    Args:
        path: absolute path to the existing schema object.
        new_path: absolute new path for the schema object.

    Raises:
        Error: If path does not exist or new_path already exists.

    Examples:
        Move a table to a different directory:

        >>>> pxt.move('dir1.my_table', 'dir2.my_table')

        Rename a table:

        >>>> pxt.move('dir1.my_table', 'dir1.new_name')
    """
    if path == new_path:
        raise excs.Error('move(): source and destination cannot be identical')
    path_obj, new_path_obj = catalog.Path(path), catalog.Path(new_path)
    if path_obj.is_ancestor(new_path_obj):
        raise excs.Error(f'move(): cannot move {path!r} into its own subdirectory')
    cat = Catalog.get()
    cat.move(path_obj, new_path_obj)


def drop_table(
    table: Union[str, catalog.Table], force: bool = False, if_not_exists: Literal['error', 'ignore'] = 'error'
) -> None:
    """Drop a table, view, or snapshot.

    Args:
        table: Fully qualified name, or handle, of the table to be dropped.
        force: If `True`, will also drop all views and sub-views of this table.
        if_not_exists: Directive regarding how to handle if the path does not exist.
            Must be one of the following:

            - `'error'`: raise an error
            - `'ignore'`: do nothing and return

    Raises:
        Error: if the qualified name

            - is invalid, or
            - does not exist and `if_not_exists='error'`, or
            - does not designate a table object, or
            - designates a table object but has dependents and `force=False`.

    Examples:
        Drop a table by its fully qualified name:
        >>> pxt.drop_table('subdir.my_table')

        Drop a table by its handle:
        >>> t = pxt.get_table('subdir.my_table')
        ... pxt.drop_table(t)

        Drop a table if it exists, otherwise do nothing:
        >>> pxt.drop_table('subdir.my_table', if_not_exists='ignore')

        Drop a table and all its dependents:
        >>> pxt.drop_table('subdir.my_table', force=True)
    """
    tbl_path: str
    if isinstance(table, catalog.Table):
        # if we're dropping a table by handle, we first need to get the current path, then drop the S lock on
        # the Table record, and then get X locks in the correct order (first containing directory, then table)
        with Catalog.get().begin_xact(for_write=False):
            tbl_path = table._path()
    else:
        assert isinstance(table, str)
        tbl_path = table

    path_obj = catalog.Path(tbl_path)
    if_not_exists_ = catalog.IfNotExistsParam.validated(if_not_exists, 'if_not_exists')
    Catalog.get().drop_table(path_obj, force=force, if_not_exists=if_not_exists_)


def list_tables(dir_path: str = '', recursive: bool = True) -> list[str]:
    """List the [`Table`][pixeltable.Table]s in a directory.

    Args:
        dir_path: Path to the directory. Defaults to the root directory.
        recursive: If `False`, returns only those tables that are directly contained in specified directory; if
            `True`, returns all tables that are descendants of the specified directory, recursively.

    Returns:
        A list of [`Table`][pixeltable.Table] paths.

    Raises:
        Error: If the path does not exist or does not designate a directory.

    Examples:
        List tables in top-level directory:

        >>> pxt.list_tables()

        List tables in 'dir1':

        >>> pxt.list_tables('dir1')
    """
    return _list_tables(dir_path, recursive=recursive, allow_system_paths=False)


def _list_tables(dir_path: str = '', recursive: bool = True, allow_system_paths: bool = False) -> list[str]:
    path_obj = catalog.Path(dir_path, empty_is_valid=True, allow_system_paths=allow_system_paths)
    contents = Catalog.get().get_dir_contents(path_obj, recursive=recursive)
    return [str(p) for p in _extract_paths(contents, parent=path_obj, entry_type=catalog.Table)]


def create_dir(
    path: str, if_exists: Literal['error', 'ignore', 'replace', 'replace_force'] = 'error', parents: bool = False
) -> Optional[catalog.Dir]:
    """Create a directory.

    Args:
        path: Path to the directory.
        if_exists: Directive regarding how to handle if the path already exists.
            Must be one of the following:

            - `'error'`: raise an error
            - `'ignore'`: do nothing and return the existing directory handle
            - `'replace'`: if the existing directory is empty, drop it and create a new one
            - `'replace_force'`: drop the existing directory and all its children, and create a new one
        parents: Create missing parent directories.

    Returns:
        A handle to the newly created directory, or to an already existing directory at the path when
            `if_exists='ignore'`. Please note the existing directory may not be empty.

    Raises:
        Error: If

            - the path is invalid, or
            - the path already exists and `if_exists='error'`, or
            - the path already exists and is not a directory, or
            - an error occurs while attempting to create the directory.

    Examples:
        >>> pxt.create_dir('my_dir')

        Create a subdirectory:

        >>> pxt.create_dir('my_dir.sub_dir')

        Create a subdirectory only if it does not already exist, otherwise do nothing:

        >>> pxt.create_dir('my_dir.sub_dir', if_exists='ignore')

        Create a directory and replace if it already exists:

        >>> pxt.create_dir('my_dir', if_exists='replace_force')

        Create a subdirectory along with its ancestors:

        >>> pxt.create_dir('parent1.parent2.sub_dir', parents=True)
    """
    path_obj = catalog.Path(path)
    if_exists_ = catalog.IfExistsParam.validated(if_exists, 'if_exists')
    return Catalog.get().create_dir(path_obj, if_exists=if_exists_, parents=parents)


def drop_dir(path: str, force: bool = False, if_not_exists: Literal['error', 'ignore'] = 'error') -> None:
    """Remove a directory.

    Args:
        path: Name or path of the directory.
        force: If `True`, will also drop all tables and subdirectories of this directory, recursively, along
            with any views or snapshots that depend on any of the dropped tables.
        if_not_exists: Directive regarding how to handle if the path does not exist.
            Must be one of the following:

            - `'error'`: raise an error
            - `'ignore'`: do nothing and return

    Raises:
        Error: If the path

            - is invalid, or
            - does not exist and `if_not_exists='error'`, or
            - is not designate a directory, or
            - is a direcotory but is not empty and `force=False`.

    Examples:
        Remove a directory, if it exists and is empty:
        >>> pxt.drop_dir('my_dir')

        Remove a subdirectory:

        >>> pxt.drop_dir('my_dir.sub_dir')

        Remove an existing directory if it is empty, but do nothing if it does not exist:

        >>> pxt.drop_dir('my_dir.sub_dir', if_not_exists='ignore')

        Remove an existing directory and all its contents:

        >>> pxt.drop_dir('my_dir', force=True)
    """
    path_obj = catalog.Path(path)  # validate format
    if_not_exists_ = catalog.IfNotExistsParam.validated(if_not_exists, 'if_not_exists')
    Catalog.get().drop_dir(path_obj, if_not_exists=if_not_exists_, force=force)


def ls(path: str = '') -> pd.DataFrame:
    """
    List the contents of a Pixeltable directory.

    This function returns a Pandas DataFrame representing a human-readable listing of the specified directory,
    including various attributes such as version and base table, as appropriate.

    To get a programmatic list of tables and/or directories, use [list_tables()][pixeltable.list_tables] and/or
    [list_dirs()][pixeltable.list_dirs] instead.
    """
    from pixeltable.catalog import retry_loop
    from pixeltable.metadata import schema

    cat = Catalog.get()
    path_obj = catalog.Path(path, empty_is_valid=True)
    dir_entries = cat.get_dir_contents(path_obj)

    @retry_loop(for_write=False)
    def op() -> list[list[str]]:
        rows: list[list[str]] = []
        for name, entry in dir_entries.items():
            if name.startswith('_'):
                continue
            if entry.dir is not None:
                kind = 'dir'
                version = ''
                base = ''
            else:
                assert entry.table is not None
                assert isinstance(entry.table, schema.Table)
                tbl = cat.get_table_by_id(entry.table.id)
                md = tbl.get_metadata()
                base = md['base'] or ''
                if base.startswith('_'):
                    base = '<anonymous base table>'
                if md['is_snapshot']:
                    kind = 'snapshot'
                elif md['is_view']:
                    kind = 'view'
                else:
                    kind = 'table'
                version = '' if kind == 'snapshot' else md['version']
                if md['is_replica']:
                    kind = f'{kind}-replica'
            rows.append([name, kind, version, base])
        return rows

    rows = op()

    rows = sorted(rows, key=lambda x: x[0])
    df = pd.DataFrame(
        {
            'Name': [row[0] for row in rows],
            'Kind': [row[1] for row in rows],
            'Version': [row[2] for row in rows],
            'Base': [row[3] for row in rows],
        },
        index=([''] * len(rows)),
    )
    return df


def _extract_paths(
    dir_entries: dict[str, Catalog.DirEntry],
    parent: catalog.Path,
    entry_type: Optional[type[catalog.SchemaObject]] = None,
) -> list[catalog.Path]:
    """Convert nested dir_entries structure to a flattened list of paths."""
    matches: list[str]
    if entry_type is None:
        matches = list(dir_entries.keys())
    elif entry_type is catalog.Dir:
        matches = [name for name, entry in dir_entries.items() if entry.dir is not None]
    else:
        matches = [name for name, entry in dir_entries.items() if entry.table is not None]

    # Filter out system paths
    matches = [name for name in matches if catalog.is_valid_identifier(name)]
    result = [parent.append(name) for name in matches]

    for name, entry in dir_entries.items():
        if len(entry.dir_entries) > 0 and catalog.is_valid_identifier(name):
            result.extend(_extract_paths(entry.dir_entries, parent=parent.append(name), entry_type=entry_type))
    return result


def list_dirs(path: str = '', recursive: bool = True) -> list[str]:
    """List the directories in a directory.

    Args:
        path: Name or path of the directory.
        recursive: If `True`, lists all descendants of this directory recursively.

    Returns:
        List of directory paths.

    Raises:
        Error: If `path_str` does not exist or does not designate a directory.

    Examples:
        >>> cl.list_dirs('my_dir', recursive=True)
        ['my_dir', 'my_dir.sub_dir1']
    """
    path_obj = catalog.Path(path, empty_is_valid=True)  # validate format
    cat = Catalog.get()
    contents = cat.get_dir_contents(path_obj, recursive=recursive)
    return [str(p) for p in _extract_paths(contents, parent=path_obj, entry_type=catalog.Dir)]


def list_functions() -> Styler:
    """Returns information about all registered functions.

    Returns:
        Pandas DataFrame with columns 'Path', 'Name', 'Parameters', 'Return Type', 'Is Agg', 'Library'
    """
    functions = func.FunctionRegistry.get().list_functions()
    paths = ['.'.join(f.self_path.split('.')[:-1]) for f in functions]
    names = [f.name for f in functions]
    params = [
        ', '.join(
            [param_name + ': ' + str(param_type) for param_name, param_type in f.signatures[0].parameters.items()]
        )
        for f in functions
    ]
    pd_df = pd.DataFrame(
        {
            'Path': paths,
            'Function Name': names,
            'Parameters': params,
            'Return Type': [str(f.signatures[0].get_return_type()) for f in functions],
        }
    )
    pd_df = pd_df.style.set_properties(None, **{'text-align': 'left'}).set_table_styles(
        [{'selector': 'th', 'props': [('text-align', 'center')]}]
    )  # center-align headings
    return pd_df.hide(axis='index')


def tools(*args: Union[func.Function, func.tools.Tool]) -> func.tools.Tools:
    """
    Specifies a collection of UDFs to be used as LLM tools. Pixeltable allows any UDF to be used as an input into an
    LLM tool-calling API. To use one or more UDFs as tools, wrap them in a `pxt.tools` call and pass the return value
    to an LLM API.

    The UDFs can be specified directly or wrapped inside a [pxt.tool()][pixeltable.tool] invocation. If a UDF is
    specified directly, the tool name will be the (unqualified) UDF name, and the tool description will consist of the
    entire contents of the UDF docstring. If a UDF is wrapped in a `pxt.tool()` invocation, then the name and/or
    description may be customized.

    Args:
        args: The UDFs to use as tools.

    Returns:
        A `Tools` instance that can be passed to an LLM tool-calling API or invoked to generate tool results.

    Examples:
        Create a tools instance with a single UDF:

        >>> tools = pxt.tools(stock_price)

        Create a tools instance with several UDFs:

        >>> tools = pxt.tools(stock_price, weather_quote)

        Create a tools instance, some of whose UDFs have customized metadata:

        >>> tools = pxt.tools(
        ...     stock_price,
        ...     pxt.tool(weather_quote, description='Returns information about the weather in a particular location.'),
        ...     pxt.tool(traffic_quote, name='traffic_conditions'),
        ... )
    """
    return func.tools.Tools(tools=[arg if isinstance(arg, func.tools.Tool) else tool(arg) for arg in args])


def tool(fn: func.Function, name: Optional[str] = None, description: Optional[str] = None) -> func.tools.Tool:
    """
    Specifies a Pixeltable UDF to be used as an LLM tool with customizable metadata. See the documentation for
    [pxt.tools()][pixeltable.tools] for more details.

    Args:
        fn: The UDF to use as a tool.
        name: The name of the tool. If not specified, then the unqualified name of the UDF will be used by default.
        description: The description of the tool. If not specified, then the entire contents of the UDF docstring
            will be used by default.

    Returns:
        A `Tool` instance that can be passed to an LLM tool-calling API.
    """
    if isinstance(fn, func.AggregateFunction):
        raise excs.Error('Aggregator UDFs cannot be used as tools')

    return func.tools.Tool(fn=fn, name=name, description=description)


def configure_logging(
    *,
    to_stdout: Optional[bool] = None,
    level: Optional[int] = None,
    add: Optional[str] = None,
    remove: Optional[str] = None,
) -> None:
    """Configure logging.

    Args:
        to_stdout: if True, also log to stdout
        level: default log level
        add: comma-separated list of 'module name:log level' pairs; ex.: add='video:10'
        remove: comma-separated list of module names
    """
    return Env.get().configure_logging(to_stdout=to_stdout, level=level, add=add, remove=remove)


def array(elements: Iterable) -> exprs.Expr:
    return exprs.Expr.from_array(elements)



from __future__ import annotations

import datetime
import io
import urllib.parse
import urllib.request
from pathlib import Path
from typing import Any, Optional

import numpy as np
import pgvector.sqlalchemy  # type: ignore[import-untyped]
import PIL
import PIL.Image
import sqlalchemy as sql

from pixeltable import env


class DataRow:
    """
    Encapsulates all data and execution state needed by RowBuilder and DataRowBatch:
    - state for in-memory computation
    - state for storing the data
    This is not meant to be a black-box abstraction.

    In-memory representations by column type:
    - StringType: str
    - IntType: int
    - FloatType: float
    - BoolType: bool
    - TimestampType: datetime.datetime
    - DateType: datetime.date
    - JsonType: json-serializable object
    - ArrayType: numpy.ndarray
    - ImageType: PIL.Image.Image
    - VideoType: local path if available, otherwise url
    - AudioType: local path if available, otherwise url
    - DocumentType: local path if available, otherwise url
    """

    vals: np.ndarray  # of object
    has_val: np.ndarray  # of bool
    excs: np.ndarray  # of object

    # If `may_have_exc` is False, then we guarantee that no slot has an exception set. This is used to optimize
    # exception handling under normal operation.
    _may_have_exc: bool

    # expr evaluation state; indexed by slot idx
    missing_slots: np.ndarray  # of bool; number of missing dependencies
    missing_dependents: np.ndarray  # of int16; number of missing dependents
    is_scheduled: np.ndarray  # of bool; True if this slot is scheduled for evaluation

    # control structures that are shared across all DataRows in a batch
    img_slot_idxs: list[int]
    media_slot_idxs: list[int]
    array_slot_idxs: list[int]

    # the primary key of a store row is a sequence of ints (the number is different for table vs view)
    pk: Optional[tuple[int, ...]]

    # file_urls:
    # - stored url of file for media in vals[i]
    # - None if vals[i] is not media type
    # - not None if file_paths[i] is not None
    file_urls: np.ndarray  # of str

    # file_paths:
    # - local path of media file in vals[i]; points to the file cache if file_urls[i] is remote
    # - None if vals[i] is not a media type or if there is no local file yet for file_urls[i]
    file_paths: np.ndarray  # of str

    # for nested rows (ie, those produced by JsonMapperDispatcher)
    parent_row: Optional[DataRow]
    parent_slot_idx: Optional[int]

    def __init__(
        self,
        size: int,
        img_slot_idxs: list[int],
        media_slot_idxs: list[int],
        array_slot_idxs: list[int],
        parent_row: Optional[DataRow] = None,
        parent_slot_idx: Optional[int] = None,
    ):
        self.img_slot_idxs = img_slot_idxs
        self.media_slot_idxs = media_slot_idxs
        self.array_slot_idxs = array_slot_idxs
        self.init(size)
        self.parent_row = parent_row
        self.parent_slot_idx = parent_slot_idx

    def init(self, num_slots: int) -> None:
        self.vals = np.full(num_slots, None, dtype=object)
        self.has_val = np.zeros(num_slots, dtype=bool)
        self.excs = np.full(num_slots, None, dtype=object)
        self._may_have_exc = False
        self.missing_slots = np.zeros(num_slots, dtype=bool)
        self.missing_dependents = np.zeros(num_slots, dtype=np.int16)
        self.is_scheduled = np.zeros(num_slots, dtype=bool)
        self.pk = None
        self.file_urls = np.full(num_slots, None, dtype=object)
        self.file_paths = np.full(num_slots, None, dtype=object)
        self.parent_row = None
        self.parent_slot_idx = None

    def clear(self, idxs: Optional[np.ndarray] = None) -> None:
        if idxs is not None:
            self.has_val[idxs] = False
            self.vals[idxs] = None
            self.excs[idxs] = None
            self.file_urls[idxs] = None
            self.file_paths[idxs] = None
        else:
            self.init(len(self.vals))

    def set_file_path(self, idx: int, path: str) -> None:
        """Augment an existing url with a local file path"""
        assert self.has_val[idx]
        assert idx in self.img_slot_idxs or idx in self.media_slot_idxs
        self.file_paths[idx] = path
        if idx in self.media_slot_idxs:
            self.vals[idx] = path

    def copy(self, target: DataRow) -> None:
        """Create a copy of the contents of this DataRow in target
        The copy shares the cell values, but not the control structures (eg, self.has_val), because these
        need to be independently updateable.
        """
        target.vals = self.vals.copy()
        target.has_val = self.has_val.copy()
        target.excs = self.excs.copy()
        target.pk = self.pk
        target.file_urls = self.file_urls.copy()
        target.file_paths = self.file_paths.copy()

    def set_pk(self, pk: tuple[int, ...]) -> None:
        self.pk = pk

    def has_exc(self, slot_idx: Optional[int] = None) -> bool:
        """
        Returns True if an exception has been set for the given slot index, or for any slot index if slot_idx is None
        """
        if not self._may_have_exc:
            return False

        if slot_idx is not None:
            return self.excs[slot_idx] is not None
        return (self.excs != None).any()

    def get_exc(self, slot_idx: int) -> Optional[Exception]:
        exc = self.excs[slot_idx]
        assert exc is None or isinstance(exc, Exception)
        return exc

    def get_first_exc(self) -> Optional[Exception]:
        mask = self.excs != None
        if not mask.any():
            return None
        return self.excs[mask][0]

    def set_exc(self, slot_idx: int, exc: Exception) -> None:
        assert self.excs[slot_idx] is None
        self.excs[slot_idx] = exc
        self._may_have_exc = True

        # an exception means the value is None
        self.has_val[slot_idx] = True
        self.vals[slot_idx] = None
        self.file_paths[slot_idx] = None
        self.file_urls[slot_idx] = None

    def __getitem__(self, index: int) -> Any:
        """Returns in-memory value, ie, what is needed for expr evaluation"""
        assert isinstance(index, int)
        if not self.has_val[index]:
            # This is a sufficiently cheap and sensitive validation that it makes sense to keep the assertion around
            # even if python is running with -O.
            raise AssertionError(index)

        if self.file_urls[index] is not None and index in self.img_slot_idxs:
            # if we need to load this from a file, it should have been materialized locally
            # TODO this fails if the url was instantiated dynamically using astype()
            assert self.file_paths[index] is not None
            if self.vals[index] is None:
                self.vals[index] = PIL.Image.open(self.file_paths[index])
                self.vals[index].load()

        return self.vals[index]

    def get_stored_val(self, index: int, sa_col_type: Optional[sql.types.TypeEngine] = None) -> Any:
        """Return the value that gets stored in the db"""
        assert self.excs[index] is None
        if not self.has_val[index]:
            # for debugging purposes
            pass
        assert self.has_val[index]

        if self.file_urls[index] is not None and (index in self.img_slot_idxs or index in self.media_slot_idxs):
            # if this is an image or other media type we want to store, we should have a url
            return self.file_urls[index]

        if self.vals[index] is not None and index in self.array_slot_idxs:
            assert isinstance(self.vals[index], np.ndarray)
            np_array = self.vals[index]
            if sa_col_type is not None and isinstance(sa_col_type, pgvector.sqlalchemy.Vector):
                return np_array
            buffer = io.BytesIO()
            np.save(buffer, np_array)
            return buffer.getvalue()

        # for JSON columns, we need to store None as an explicit NULL, otherwise it stores a json 'null'
        if self.vals[index] is None and sa_col_type is not None and isinstance(sa_col_type, sql.JSON):
            return sql.sql.null()

        if isinstance(self.vals[index], datetime.datetime) and self.vals[index].tzinfo is None:
            # if the datetime is naive, cast it to the default time zone
            return self.vals[index].replace(tzinfo=env.Env.get().default_time_zone)

        return self.vals[index]

    def __setitem__(self, idx: int, val: Any) -> None:
        """Assign in-memory cell value
        This allows overwriting
        """
        assert isinstance(idx, int)
        assert self.excs[idx] is None

        if (idx in self.img_slot_idxs or idx in self.media_slot_idxs) and isinstance(val, str):
            # this is either a local file path or a URL
            parsed = urllib.parse.urlparse(val)
            # Determine if this is a local file or a remote URL. If the scheme length is <= 1,
            # we assume it's a local file. (This is because a Windows path will be interpreted
            # by urllib as a URL with scheme equal to the drive letter.)
            if len(parsed.scheme) <= 1 or parsed.scheme == 'file':
                # local file path
                assert self.file_urls[idx] is None and self.file_paths[idx] is None
                if len(parsed.scheme) <= 1:
                    path = str(Path(val).absolute())  # Ensure we're using an absolute pathname.
                    self.file_urls[idx] = urllib.parse.urljoin('file:', urllib.request.pathname2url(path))
                    self.file_paths[idx] = path
                else:  # file:// URL
                    self.file_urls[idx] = val
                    # Wrap the path in a url2pathname() call to ensure proper handling on Windows.
                    self.file_paths[idx] = urllib.parse.unquote(urllib.request.url2pathname(parsed.path))
            else:
                # URL
                assert self.file_urls[idx] is None
                self.file_urls[idx] = val

            if idx in self.media_slot_idxs:
                self.vals[idx] = self.file_paths[idx] if self.file_paths[idx] is not None else self.file_urls[idx]
        elif idx in self.array_slot_idxs and isinstance(val, bytes):
            self.vals[idx] = np.load(io.BytesIO(val))
        else:
            self.vals[idx] = val
        self.has_val[idx] = True

    def flush_img(self, index: int, filepath: Optional[str] = None) -> None:
        """Discard the in-memory value and save it to a local file, if filepath is not None"""
        if self.vals[index] is None:
            return
        assert self.excs[index] is None
        if self.file_paths[index] is None:
            if filepath is not None:
                image = self.vals[index]
                assert isinstance(image, PIL.Image.Image)
                # Default to JPEG unless the image has a transparency layer (which isn't supported by JPEG).
                # In that case, use WebP instead.
                format = 'webp' if image.has_transparency_data else 'jpeg'
                if not filepath.endswith(f'.{format}'):
                    filepath += f'.{format}'
                self.file_paths[index] = filepath
                self.file_urls[index] = urllib.parse.urljoin('file:', urllib.request.pathname2url(filepath))
                image.save(filepath, format=format)
            else:
                # we discard the content of this cell
                self.has_val[index] = False
        else:
            # we already have a file for this image, nothing left to do
            pass
        self.vals[index] = None

    @property
    def rowid(self) -> tuple[int, ...]:
        return self.pk[:-1]

    @property
    def v_min(self) -> int:
        return self.pk[-1]



import logging
from typing import TYPE_CHECKING

import PIL.Image

import pixeltable as pxt
from pixeltable.func import Batch
from pixeltable.functions.util import normalize_image_mode
from pixeltable.utils.code import local_public_names

if TYPE_CHECKING:
    from yolox.models import Yolox, YoloxProcessor  # type: ignore[import-untyped]

_logger = logging.getLogger('pixeltable')


@pxt.udf(batch_size=4)
def yolox(images: Batch[PIL.Image.Image], *, model_id: str, threshold: float = 0.5) -> Batch[dict]:
    """
    Computes YOLOX object detections for the specified image. `model_id` should reference one of the models
    defined in the [YOLOX documentation](https://github.com/Megvii-BaseDetection/YOLOX).

    YOLOX is part of the `pixeltable.ext` package: long-term support in Pixeltable is not guaranteed.

    __Requirements__:

    - `pip install pixeltable-yolox`

    Args:
        model_id: one of: `yolox_nano`, `yolox_tiny`, `yolox_s`, `yolox_m`, `yolox_l`, `yolox_x`
        threshold: the threshold for object detection

    Returns:
        A dictionary containing the output of the object detection model.

    Examples:
        Add a computed column that applies the model `yolox_m` to an existing
        Pixeltable column `tbl.image` of the table `tbl`:

        >>> tbl.add_computed_column(detections=yolox(tbl.image, model_id='yolox_m', threshold=0.8))
    """
    import torch

    model = _lookup_model(model_id, 'cpu')
    processor = _lookup_processor(model_id)
    normalized_images = [normalize_image_mode(image) for image in images]
    with torch.no_grad():
        tensor = processor(normalized_images)
        output = model(tensor)
        return processor.postprocess(normalized_images, output, threshold=threshold)


@pxt.udf
def yolo_to_coco(detections: dict) -> list:
    """
    Converts the output of a YOLOX object detection model to COCO format.

    YOLOX is part of the `pixeltable.ext` package: long-term support in Pixeltable is not guaranteed.

    Args:
        detections: The output of a YOLOX object detection model, as returned by `yolox`.

    Returns:
        A dictionary containing the data from `detections`, converted to COCO format.

    Examples:
        Add a computed column that converts the output `tbl.detections` to COCO format, where `tbl.image`
        is the image for which detections were computed:

        >>> tbl.add_computed_column(detections=yolox(tbl.image, model_id='yolox_m', threshold=0.8))
        ... tbl.add_computed_column(detections_coco=yolo_to_coco(tbl.detections))
    """
    bboxes, labels = detections['bboxes'], detections['labels']
    num_annotations = len(detections['bboxes'])
    assert num_annotations == len(detections['labels'])
    result = []
    for i in range(num_annotations):
        bbox = bboxes[i]
        ann = {
            'bbox': [round(bbox[0]), round(bbox[1]), round(bbox[2] - bbox[0]), round(bbox[3] - bbox[1])],
            'category': labels[i],
        }
        result.append(ann)
    return result


def _lookup_model(model_id: str, device: str) -> 'Yolox':
    from yolox.models import Yolox

    key = (model_id, device)
    if key not in _model_cache:
        _model_cache[key] = Yolox.from_pretrained(model_id, device=device)

    return _model_cache[key]


def _lookup_processor(model_id: str) -> 'YoloxProcessor':
    from yolox.models import YoloxProcessor

    if model_id not in _processor_cache:
        _processor_cache[model_id] = YoloxProcessor(model_id)

    return _processor_cache[model_id]


_model_cache: dict[tuple[str, str], 'Yolox'] = {}
_processor_cache: dict[str, 'YoloxProcessor'] = {}


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



from __future__ import annotations

import dataclasses
import json
import logging
from uuid import UUID

import sqlalchemy as sql

from pixeltable.env import Env
from pixeltable.metadata import schema

from .schema_object import SchemaObject

_logger = logging.getLogger('pixeltable')


class Dir(SchemaObject):
    def __init__(self, id: UUID, parent_id: UUID, name: str):
        super().__init__(id, name, parent_id)

    @classmethod
    def _create(cls, parent_id: UUID, name: str) -> Dir:
        session = Env.get().session
        user = Env.get().user
        assert session is not None
        dir_md = schema.DirMd(name=name, user=user, additional_md={})
        dir_record = schema.Dir(parent_id=parent_id, md=dataclasses.asdict(dir_md))
        session.add(dir_record)
        session.flush()
        # print(f'{datetime.datetime.now()} create dir {dir_record}')
        assert dir_record.id is not None
        assert isinstance(dir_record.id, UUID)
        dir = cls(dir_record.id, parent_id, name)
        return dir

    def _display_name(self) -> str:
        return 'directory'

    def _path(self) -> str:
        """Returns the path to this schema object."""
        if self._dir_id is None:
            # we're the root dir
            return ''
        return super()._path()

    def _move(self, new_name: str, new_dir_id: UUID) -> None:
        # print(
        #     f'{datetime.datetime.now()} move dir name={self._name} parent={self._dir_id} '
        #     f'new_name={new_name} new_dir_id={new_dir_id}'
        # )
        super()._move(new_name, new_dir_id)
        stmt = sql.text(
            (
                f'UPDATE {schema.Dir.__table__} '
                f'SET {schema.Dir.parent_id.name} = :new_dir_id, '
                f"    {schema.Dir.md.name} = jsonb_set({schema.Dir.md.name}, '{{name}}', (:new_name)::jsonb) "
                f'WHERE {schema.Dir.id.name} = :id'
            )
        )
        Env.get().conn.execute(stmt, {'new_dir_id': new_dir_id, 'new_name': json.dumps(new_name), 'id': self._id})



from typing import Any, Optional

import sqlalchemy as sql

from pixeltable import exprs
from pixeltable.metadata import register_converter
from pixeltable.metadata.converters.util import convert_table_md


@register_converter(version=29)
def _(engine: sql.engine.Engine) -> None:
    convert_table_md(engine, substitution_fn=__substitute_md)


def __substitute_md(k: Optional[str], v: Any) -> Optional[tuple[Optional[str], Any]]:
    # Defaults are now stored as literals in signatures
    if k == 'parameters':
        for param in v:
            assert isinstance(param, dict)
            has_default = param.get('has_default') or (param.get('default') is not None)
            if 'has_default' in param:
                del param['has_default']
            literal = exprs.Expr.from_object(param['default']) if has_default else None
            assert literal is None or isinstance(literal, exprs.Literal)
            param['default'] = None if literal is None else literal.as_dict()
        return k, v

    # Method of organizing argument expressions has changed
    if isinstance(v, dict) and v.get('_classname') == 'FunctionCall':
        args = v['args']
        kwargs = v['kwargs']
        components = v['components']
        group_by_start_idx = v['group_by_start_idx']
        group_by_stop_idx = v['group_by_stop_idx']
        order_by_start_idx = v['order_by_start_idx']

        new_args = []
        for arg in args:
            if arg[0] is not None:
                assert isinstance(arg[0], int)
                new_args.append(components[arg[0]])
            else:
                literal = exprs.Expr.from_object(arg[1])
                new_args.append(literal.as_dict())

        new_kwargs = {}
        for name, kwarg in kwargs.items():
            if kwarg[0] is not None:
                assert isinstance(kwarg[0], int)
                new_kwargs[name] = components[kwarg[0]]
            else:
                literal = exprs.Expr.from_object(kwarg[1])
                new_kwargs[name] = literal.as_dict()

        # We need to expand ("unroll") any var-args or var-kwargs.

        new_args_len = len(new_args)
        rolled_args: Optional[dict] = None
        rolled_kwargs: Optional[dict] = None

        if 'signature' in v['fn']:
            # If it's a pickled function, there's no signature, so we're out of luck; varargs in a pickled function
            # is an edge case that won't migrate properly.
            parameters: list[dict] = v['fn']['signature']['parameters']
            for i, param in enumerate(parameters):
                if param['kind'] == 'VAR_POSITIONAL' and new_args_len > i:
                    # For peculiar historical reasons, variable kwargs might show up in args. Thus variable
                    # positional args is not necessarily the last element of args; it might be the second-to-last.
                    assert new_args_len <= i + 2, new_args
                    rolled_args = new_args[i]
                    new_args = new_args[:i] + new_args[i + 1 :]
                if param['kind'] == 'VAR_KEYWORD':
                    # As noted above, variable kwargs might show up either in args or in kwargs. If it's in args, it
                    # is necessarily the last element.
                    if new_args_len > i:
                        assert new_args_len <= i + 1, new_args
                        rolled_kwargs = new_args.pop()
                    if param['name'] in kwargs:
                        assert rolled_kwargs is None
                        rolled_kwargs = kwargs.pop(param['name'])

        if rolled_args is not None:
            assert rolled_args['_classname'] in ('InlineArray', 'InlineList')
            new_args.extend(rolled_args['components'])
        if rolled_kwargs is not None:
            assert rolled_kwargs['_classname'] == 'InlineDict'
            new_kwargs.update(zip(rolled_kwargs['keys'], rolled_kwargs['components']))

        group_by_exprs = [components[i] for i in range(group_by_start_idx, group_by_stop_idx)]
        order_by_exprs = [components[i] for i in range(order_by_start_idx, len(components))]

        new_components = [*new_args, *new_kwargs.values(), *group_by_exprs, *order_by_exprs]

        newv = {
            'fn': v['fn'],
            'arg_idxs': list(range(len(new_args))),
            'kwarg_idxs': {name: i + len(new_args) for i, name in enumerate(new_kwargs.keys())},
            'group_by_start_idx': len(new_args) + len(new_kwargs),
            'group_by_stop_idx': len(new_args) + len(new_kwargs) + len(group_by_exprs),
            'order_by_start_idx': len(new_args) + len(new_kwargs) + len(group_by_exprs),
            'is_method_call': False,
            '_classname': 'FunctionCall',
            'components': new_components,
        }
        if 'return_type' in v:
            newv['return_type'] = v['return_type']

        return k, newv

    return None



from __future__ import annotations

import logging
from typing import Iterator

from pixeltable import exceptions as excs

from .globals import is_valid_path

_logger = logging.getLogger('pixeltable')


class Path:
    def __init__(self, path: str, empty_is_valid: bool = False, allow_system_paths: bool = False):
        if not is_valid_path(path, empty_is_valid, allow_system_paths):
            raise excs.Error(f"Invalid path format: '{path}'")
        self.components = path.split('.')

    @property
    def len(self) -> int:
        return 0 if self.is_root else len(self.components)

    @property
    def name(self) -> str:
        assert len(self.components) > 0
        return self.components[-1]

    @property
    def is_root(self) -> bool:
        return not self.components[0]

    @property
    def is_system_path(self) -> bool:
        return self.components[0].startswith('_')

    @property
    def parent(self) -> Path:
        if len(self.components) == 1:
            if self.is_root:
                return self
            else:
                return Path('', empty_is_valid=True, allow_system_paths=True)
        else:
            return Path('.'.join(self.components[:-1]), allow_system_paths=True)

    def append(self, name: str) -> Path:
        if self.is_root:
            return Path(name, allow_system_paths=True)
        else:
            return Path(f'{self}.{name}', allow_system_paths=True)

    def is_ancestor(self, other: Path, is_parent: bool = False) -> bool:
        """
        True if self as an ancestor path of other.
        """
        if self.len >= other.len or other.is_root:
            return False
        if self.is_root and (other.len == 1 or not is_parent):
            return True
        is_prefix = self.components == other.components[: self.len]
        return is_prefix and (self.len == (other.len - 1) or not is_parent)

    def ancestors(self) -> Iterator[Path]:
        """
        Return all ancestors of this path in top-down order including root.
        If this path is for the root directory, which has no parent, then None is returned.
        """
        if self.is_root:
            return
        else:
            for i in range(0, len(self.components)):
                yield Path('.'.join(self.components[0:i]), empty_is_valid=True)

    def __repr__(self) -> str:
        return repr(str(self))

    def __str__(self) -> str:
        return '.'.join(self.components)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Path) and str(self) == str(other)

    def __hash__(self) -> int:
        return hash(str(self))

    def __lt__(self, other: Path) -> bool:
        return str(self) < str(other)



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs) for `StringType`.
It closely follows the Pandas `pandas.Series.str` API.

Example:
```python
import pixeltable as pxt

t = pxt.get_table(...)
t.select(t.str_col.capitalize()).collect()
```
"""

import builtins
import re
import textwrap
from string import whitespace
from typing import Any, Optional

import sqlalchemy as sql

import pixeltable as pxt
from pixeltable.utils.code import local_public_names


@pxt.udf(is_method=True)
def capitalize(self: str) -> str:
    """
    Return string with its first character capitalized and the rest lowercased.

    Equivalent to [`str.capitalize()`](https://docs.python.org/3/library/stdtypes.html#str.capitalize).
    """
    return self.capitalize()


@capitalize.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.func.concat(sql.func.upper(sql.func.left(self, 1)), sql.func.lower(sql.func.right(self, -1)))


@pxt.udf(is_method=True)
def casefold(self: str) -> str:
    """
    Return a casefolded copy of string.

    Equivalent to [`str.casefold()`](https://docs.python.org/3/library/stdtypes.html#str.casefold).
    """
    return self.casefold()


@pxt.udf(is_method=True)
def center(self: str, width: int, fillchar: str = ' ') -> str:
    """
    Return a centered string of length `width`.

    Equivalent to [`str.center()`](https://docs.python.org/3/library/stdtypes.html#str.center).

    Args:
        width: Total width of the resulting string.
        fillchar: Character used for padding.
    """
    return self.center(width, fillchar)


@pxt.udf(is_method=True)
def contains(self: str, substr: str, case: bool = True) -> bool:
    """
    Test if string contains a substring.

    Args:
        substr: string literal or regular expression
        case: if False, ignore case
    """
    if case:
        return substr in self
    else:
        return substr.lower() in self.lower()


@contains.to_sql
def _(
    self: sql.ColumnElement, substr: sql.ColumnElement, case: Optional[sql.ColumnElement] = None
) -> sql.ColumnElement:
    # Replace all occurrences of `%`, `_`, and `\` with escaped versions
    escaped_substr = sql.func.regexp_replace(substr, r'(%|_|\\)', r'\\\1', 'g')
    if case is None:
        # Default `case` is True, so we do a case-sensitive comparison
        return self.like(sql.func.concat('%', escaped_substr, '%'))
    else:
        # Toggle case-sensitivity based on the value of `case`
        return sql.case(
            (case, self.like(sql.func.concat('%', escaped_substr, '%'))),
            else_=sql.func.lower(self).like(sql.func.concat('%', sql.func.lower(escaped_substr), '%')),
        )


@pxt.udf(is_method=True)
def contains_re(self: str, pattern: str, flags: int = 0) -> bool:
    """
    Test if string contains a regular expression pattern.

    Args:
        pattern: regular expression pattern
        flags: [flags](https://docs.python.org/3/library/re.html#flags) for the `re` module
    """
    return bool(re.search(pattern, self, flags))


@pxt.udf(is_method=True)
def count(self: str, pattern: str, flags: int = 0) -> int:
    """
    Count occurrences of pattern or regex.

    Args:
        pattern: string literal or regular expression
        flags: [flags](https://docs.python.org/3/library/re.html#flags) for the `re` module
    """
    return builtins.len(re.findall(pattern, self, flags))


@pxt.udf(is_method=True)
def endswith(self: str, substr: str) -> bool:
    """
    Return `True` if the string ends with the specified suffix, otherwise return `False`.

    Equivalent to [`str.endswith()`](https://docs.python.org/3/library/stdtypes.html#str.endswith).

    Args:
        substr: string literal
    """
    return self.endswith(substr)


@endswith.to_sql
def _(self: sql.ColumnElement, substr: sql.ColumnElement) -> sql.ColumnElement:
    # Replace all occurrences of `%`, `_`, and `\` with escaped versions
    escaped_substr = sql.func.regexp_replace(substr, r'(%|_|\\)', r'\\\1', 'g')
    return self.like(sql.func.concat('%', escaped_substr))


@pxt.udf(is_method=True)
def fill(self: str, width: int, **kwargs: Any) -> str:
    """
    Wraps the single paragraph in string, and returns a single string containing the wrapped paragraph.

    Equivalent to [`textwrap.fill()`](https://docs.python.org/3/library/textwrap.html#textwrap.fill).

    Args:
        width: Maximum line width.
        kwargs: Additional keyword arguments to pass to `textwrap.fill()`.
    """
    return textwrap.fill(self, width, **kwargs)


@pxt.udf(is_method=True)
def find(self: str, substr: str, start: int = 0, end: Optional[int] = None) -> int:
    """
    Return the lowest index in string where `substr` is found within the slice `s[start:end]`.

    Equivalent to [`str.find()`](https://docs.python.org/3/library/stdtypes.html#str.find).

    Args:
        substr: substring to search for
        start: slice start
        end: slice end
    """
    return self.find(substr, start, end)


@find.to_sql
def _(
    self: sql.ColumnElement,
    substr: sql.ColumnElement,
    start: sql.ColumnElement,
    end: Optional[sql.ColumnElement] = None,
) -> sql.ColumnElement:
    sl = pxt.functions.string.slice._to_sql(self, start, end)
    if sl is None:
        return None

    strpos = sql.func.strpos(sl, substr)
    return sql.case(
        (strpos == 0, -1), (start >= 0, strpos + start - 1), else_=strpos + sql.func.char_length(self) + start - 1
    )


@pxt.udf(is_method=True)
def findall(self: str, pattern: str, flags: int = 0) -> list:
    """
    Find all occurrences of a regular expression pattern in string.

    Equivalent to [`re.findall()`](https://docs.python.org/3/library/re.html#re.findall).

    Args:
        pattern: regular expression pattern
        flags: [flags](https://docs.python.org/3/library/re.html#flags) for the `re` module
    """
    return re.findall(pattern, self, flags)


@pxt.udf(is_method=True)
def format(self: str, *args: Any, **kwargs: Any) -> str:
    """
    Perform string formatting.

    Equivalent to [`str.format()`](https://docs.python.org/3/library/stdtypes.html#str.format).
    """
    return self.format(*args, **kwargs)


@pxt.udf(is_method=True)
def fullmatch(self: str, pattern: str, case: bool = True, flags: int = 0) -> bool:
    """
    Determine if string fully matches a regular expression.

    Equivalent to [`re.fullmatch()`](https://docs.python.org/3/library/re.html#re.fullmatch).

    Args:
        pattern: regular expression pattern
        case: if False, ignore case
        flags: [flags](https://docs.python.org/3/library/re.html#flags) for the `re` module
    """
    if not case:
        flags |= re.IGNORECASE
    _ = bool(re.fullmatch(pattern, self, flags))
    return bool(re.fullmatch(pattern, self, flags))


@pxt.udf(is_method=True)
def index(self: str, substr: str, start: int = 0, end: Optional[int] = None) -> int:
    """
    Return the lowest index in string where `substr` is found within the slice `[start:end]`.
    Raises ValueError if `substr` is not found.

    Equivalent to [`str.index()`](https://docs.python.org/3/library/stdtypes.html#str.index).

    Args:
        substr: substring to search for
        start: slice start
        end: slice end
    """
    return self.index(substr, start, end)


@pxt.udf(is_method=True)
def isalnum(self: str) -> bool:
    """
    Return `True` if all characters in the string are alphanumeric and there is at least one character, `False`
    otherwise.

    Equivalent to [`str.isalnum()`](https://docs.python.org/3/library/stdtypes.html#str.isalnum
    """
    return self.isalnum()


@pxt.udf(is_method=True)
def isalpha(self: str) -> bool:
    """
    Return `True` if all characters in the string are alphabetic and there is at least one character, `False` otherwise.

    Equivalent to [`str.isalpha()`](https://docs.python.org/3/library/stdtypes.html#str.isalpha).
    """
    return self.isalpha()


@pxt.udf(is_method=True)
def isascii(self: str) -> bool:
    """
    Return `True` if the string is empty or all characters in the string are ASCII, `False` otherwise.

    Equivalent to [`str.isascii()`](https://docs.python.org/3/library/stdtypes.html#str.isascii).
    """
    return self.isascii()


@pxt.udf(is_method=True)
def isdecimal(self: str) -> bool:
    """
    Return `True` if all characters in the string are decimal characters and there is at least one character, `False`
    otherwise.

    Equivalent to [`str.isdecimal()`](https://docs.python.org/3/library/stdtypes.html#str.isdecimal).
    """
    return self.isdecimal()


@pxt.udf(is_method=True)
def isdigit(self: str) -> bool:
    """
    Return `True` if all characters in the string are digits and there is at least one character, `False` otherwise.

    Equivalent to [`str.isdigit()`](https://docs.python.org/3/library/stdtypes.html#str.isdigit).
    """
    return self.isdigit()


@pxt.udf(is_method=True)
def isidentifier(self: str) -> bool:
    """
    Return `True` if the string is a valid identifier according to the language definition, `False` otherwise.

    Equivalent to [`str.isidentifier()`](https://docs.python.org/3/library/stdtypes.html#str.isidentifier)
    """
    return self.isidentifier()


@pxt.udf(is_method=True)
def islower(self: str) -> bool:
    """
    Return `True` if all cased characters in the string are lowercase and there is at least one cased character,
    `False` otherwise.

    Equivalent to [`str.islower()`](https://docs.python.org/3/library/stdtypes.html#str.islower)
    """
    return self.islower()


@pxt.udf(is_method=True)
def isnumeric(self: str) -> bool:
    """
    Return `True` if all characters in the string are numeric characters, `False` otherwise.

    Equivalent to [`str.isnumeric()`](https://docs.python.org/3/library/stdtypes.html#str.isnumeric)
    """
    return self.isnumeric()


@pxt.udf(is_method=True)
def isupper(self: str) -> bool:
    """
    Return `True` if all cased characters in the string are uppercase and there is at least one cased character,
    `False` otherwise.

    Equivalent to [`str.isupper()`](https://docs.python.org/3/library/stdtypes.html#str.isupper)
    """
    return self.isupper()


@pxt.udf(is_method=True)
def istitle(self: str) -> bool:
    """
    Return `True` if the string is a titlecased string and there is at least one character, `False` otherwise.

    Equivalent to [`str.istitle()`](https://docs.python.org/3/library/stdtypes.html#str.istitle)
    """
    return self.istitle()


@pxt.udf(is_method=True)
def isspace(self: str) -> bool:
    """
    Return `True` if there are only whitespace characters in the string and there is at least one character,
    `False` otherwise.

    Equivalent to [`str.isspace()`](https://docs.python.org/3/library/stdtypes.html#str.isspace)
    """
    return self.isspace()


@pxt.udf
def join(sep: str, elements: list) -> str:
    """
    Return a string which is the concatenation of the strings in `elements`.

    Equivalent to [`str.join()`](https://docs.python.org/3/library/stdtypes.html#str.join)
    """
    return sep.join(elements)


@pxt.udf(is_method=True)
def len(self: str) -> int:
    """
    Return the number of characters in the string.

    Equivalent to [`len(str)`](https://docs.python.org/3/library/functions.html#len)
    """
    return builtins.len(self)


@len.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.func.char_length(self)


@pxt.udf(is_method=True)
def ljust(self: str, width: int, fillchar: str = ' ') -> str:
    """
    Return the string left-justified in a string of length `width`.

    Equivalent to [`str.ljust()`](https://docs.python.org/3/library/stdtypes.html#str.ljust)

    Args:
        width: Minimum width of resulting string; additional characters will be filled with character defined in
            `fillchar`.
        fillchar: Additional character for filling.
    """
    return self.ljust(width, fillchar)


@pxt.udf(is_method=True)
def lower(self: str) -> str:
    """
    Return a copy of the string with all the cased characters converted to lowercase.

    Equivalent to [`str.lower()`](https://docs.python.org/3/library/stdtypes.html#str.lower)
    """
    return self.lower()


@lower.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.func.lower(self)


@pxt.udf(is_method=True)
def lstrip(self: str, chars: Optional[str] = None) -> str:
    """
    Return a copy of the string with leading characters removed. The `chars` argument is a string specifying the set of
    characters to be removed. If omitted or `None`, whitespace characters are removed.

    Equivalent to [`str.lstrip()`](https://docs.python.org/3/library/stdtypes.html#str.lstrip)

    Args:
        chars: The set of characters to be removed.
    """
    return self.lstrip(chars)


@lstrip.to_sql
def _(self: sql.ColumnElement, chars: Optional[sql.ColumnElement] = None) -> sql.ColumnElement:
    return sql.func.ltrim(self, chars if chars is not None else whitespace)


@pxt.udf(is_method=True)
def match(self: str, pattern: str, case: bool = True, flags: int = 0) -> bool:
    """
    Determine if string starts with a match of a regular expression

    Args:
        pattern: regular expression pattern
        case: if False, ignore case
        flags: [flags](https://docs.python.org/3/library/re.html#flags) for the `re` module
    """
    if not case:
        flags |= re.IGNORECASE
    return bool(re.match(pattern, self, flags))


@pxt.udf(is_method=True)
def normalize(self: str, form: str) -> str:
    """
    Return the Unicode normal form.

    Equivalent to [`unicodedata.normalize()`](https://docs.python.org/3/library/unicodedata.html#unicodedata.normalize)

    Args:
        form: Unicode normal form (`'NFC'`, `'NFKC'`, `'NFD'`, `'NFKD'`)
    """
    import unicodedata

    return unicodedata.normalize(form, self)  # type: ignore[arg-type]


@pxt.udf(is_method=True)
def pad(self: str, width: int, side: str = 'left', fillchar: str = ' ') -> str:
    """
    Pad string up to width

    Args:
        width: Minimum width of resulting string; additional characters will be filled with character defined in
            `fillchar`.
        side: Side from which to fill resulting string (`'left'`, `'right'`, `'both'`)
        fillchar: Additional character for filling
    """
    if side == 'left':
        return self.ljust(width, fillchar)
    elif side == 'right':
        return self.rjust(width, fillchar)
    elif side == 'both':
        return self.center(width, fillchar)
    else:
        raise ValueError(f'Invalid side: {side}')


@pxt.udf(is_method=True)
def partition(self: str, sep: str = ' ') -> list:
    """
    Splits string at the first occurrence of `sep`, and returns 3 elements containing the part before the
    separator, the separator itself, and the part after the separator. If the separator is not found, return 3 elements
    containing string itself, followed by two empty strings.
    """
    idx = self.find(sep)
    if idx == -1:
        return [self, '', '']
    return [self[:idx], sep, self[idx + builtins.len(sep) :]]


@pxt.udf(is_method=True)
def removeprefix(self: str, prefix: str) -> str:
    """
    Remove prefix. If the prefix is not present, returns string.
    """
    return self.removeprefix(prefix)


@removeprefix.to_sql
def _(self: sql.ColumnElement, prefix: sql.ColumnElement) -> sql.ColumnElement:
    return sql.case((startswith._to_sql(self, prefix), sql.func.right(self, -sql.func.char_length(prefix))), else_=self)


@pxt.udf(is_method=True)
def removesuffix(self: str, suffix: str) -> str:
    """
    Remove suffix. If the suffix is not present, returns string.
    """
    return self.removesuffix(suffix)


@removesuffix.to_sql
def _(self: sql.ColumnElement, suffix: sql.ColumnElement) -> sql.ColumnElement:
    return sql.case((endswith._to_sql(self, suffix), sql.func.left(self, -sql.func.char_length(suffix))), else_=self)


@pxt.udf(is_method=True)
def repeat(self: str, n: int) -> str:
    """
    Repeat string `n` times.
    """
    return self * n


@repeat.to_sql
def _(self: sql.ColumnElement, n: sql.ColumnElement) -> sql.ColumnElement:
    return sql.func.repeat(self, n.cast(sql.types.INT))


@pxt.udf(is_method=True)
def replace(self: str, substr: str, repl: str, n: Optional[int] = None) -> str:
    """
    Replace occurrences of `substr` with `repl`.

    Equivalent to [`str.replace()`](https://docs.python.org/3/library/stdtypes.html#str.replace).

    Args:
        substr: string literal
        repl: replacement string
        n: number of replacements to make (if `None`, replace all occurrences)
    """
    return self.replace(substr, repl, n or -1)


@replace.to_sql
def _(
    self: sql.ColumnElement, substr: sql.ColumnElement, repl: sql.ColumnElement, n: Optional[sql.ColumnElement] = None
) -> sql.ColumnElement:
    if n is not None:
        return None  # SQL does not support bounding the number of replacements

    return sql.func.replace(self, substr, repl)


@pxt.udf(is_method=True)
def replace_re(self: str, pattern: str, repl: str, n: Optional[int] = None, flags: int = 0) -> str:
    """
    Replace occurrences of a regular expression pattern with `repl`.

    Equivalent to [`re.sub()`](https://docs.python.org/3/library/re.html#re.sub).

    Args:
        pattern: regular expression pattern
        repl: replacement string
        n: number of replacements to make (if `None`, replace all occurrences)
        flags: [flags](https://docs.python.org/3/library/re.html#flags) for the `re` module
    """
    return re.sub(pattern, repl, self, count=(n or 0), flags=flags)


@pxt.udf(is_method=True)
def reverse(self: str) -> str:
    """
    Return a reversed copy of the string.

    Equivalent to `str[::-1]`.
    """
    return self[::-1]


@reverse.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.func.reverse(self)


@pxt.udf(is_method=True)
def rfind(self: str, substr: str, start: Optional[int] = 0, end: Optional[int] = None) -> int:
    """
    Return the highest index where `substr` is found, such that `substr` is contained within `[start:end]`.

    Equivalent to [`str.rfind()`](https://docs.python.org/3/library/stdtypes.html#str.rfind).

    Args:
        substr: substring to search for
        start: slice start
        end: slice end
    """
    return self.rfind(substr, start, end)


@pxt.udf(is_method=True)
def rindex(self: str, substr: str, start: Optional[int] = 0, end: Optional[int] = None) -> int:
    """
    Return the highest index where `substr` is found, such that `substr` is contained within `[start:end]`.
    Raises ValueError if `substr` is not found.

    Equivalent to [`str.rindex()`](https://docs.python.org/3/library/stdtypes.html#str.rindex).
    """
    return self.rindex(substr, start, end)


@pxt.udf(is_method=True)
def rjust(self: str, width: int, fillchar: str = ' ') -> str:
    """
    Return the string right-justified in a string of length `width`.

    Equivalent to [`str.rjust()`](https://docs.python.org/3/library/stdtypes.html#str.rjust).

    Args:
        width: Minimum width of resulting string.
        fillchar: Additional character for filling.
    """
    return self.rjust(width, fillchar)


@pxt.udf(is_method=True)
def rpartition(self: str, sep: str = ' ') -> list:
    """
    This method splits string at the last occurrence of `sep`, and returns a list containing the part before the
    separator, the separator itself, and the part after the separator.
    """
    idx = self.rfind(sep)
    if idx == -1:
        return [self, '', '']
    return [self[:idx], sep, self[idx + builtins.len(sep) :]]


@pxt.udf(is_method=True)
def rstrip(self: str, chars: Optional[str] = None) -> str:
    """
    Return a copy of string with trailing characters removed.

    Equivalent to [`str.rstrip()`](https://docs.python.org/3/library/stdtypes.html#str.rstrip).

    Args:
        chars: The set of characters to be removed. If omitted or `None`, whitespace characters are removed.
    """
    return self.rstrip(chars)


@rstrip.to_sql
def _(self: sql.ColumnElement, chars: Optional[sql.ColumnElement] = None) -> sql.ColumnElement:
    return sql.func.rtrim(self, chars if chars is not None else whitespace)


@pxt.udf(is_method=True)
def slice(self: str, start: Optional[int] = None, stop: Optional[int] = None, step: Optional[int] = None) -> str:
    """
    Return a slice.

    Args:
        start: slice start
        stop: slice end
        step: slice step
    """
    return self[start:stop:step]


@slice.to_sql
def _(
    self: sql.ColumnElement,
    start: Optional[sql.ColumnElement] = None,
    stop: Optional[sql.ColumnElement] = None,
    step: Optional[sql.ColumnElement] = None,
) -> sql.ColumnElement:
    if step is not None:
        return None

    if start is not None:
        start = start.cast(sql.types.INT)  # Postgres won't accept a BIGINT
        start = sql.case(
            (start >= 0, start + 1),  # SQL is 1-based, Python is 0-based
            else_=sql.func.char_length(self) + start + 1,  # negative index
        )
        start = sql.func.greatest(start, 1)

    if stop is not None:
        stop = stop.cast(sql.types.INT)  # Postgres won't accept a BIGINT
        stop = sql.case(
            (stop >= 0, stop + 1),  # SQL is 1-based, Python is 0-based
            else_=sql.func.char_length(self) + stop + 1,  # negative index
        )
        stop = sql.func.greatest(stop, 0)

    if start is None:
        if stop is None:
            return self
        return sql.func.substr(self, 1, stop)
    if stop is None:
        return sql.func.substr(self, start)
    return sql.func.substr(self, start, sql.func.greatest(stop - start, 0))


@pxt.udf(is_method=True)
def slice_replace(
    self: str, start: Optional[int] = None, stop: Optional[int] = None, repl: Optional[str] = None
) -> str:
    """
    Replace a positional slice with another value.

    Args:
        start: slice start
        stop: slice end
        repl: replacement value
    """
    return self[:start] + repl + self[stop:]


@pxt.udf(is_method=True)
def startswith(self: str, substr: str) -> int:
    """
    Return `True` if string starts with `substr`, otherwise return `False`.

    Equivalent to [`str.startswith()`](https://docs.python.org/3/library/stdtypes.html#str.startswith).

    Args:
        substr: string literal
    """
    return self.startswith(substr)


@startswith.to_sql
def _(self: sql.ColumnElement, substr: sql.ColumnElement) -> sql.ColumnElement:
    # Replace all occurrences of `%`, `_`, and `\` with escaped versions
    escaped_substr = sql.func.regexp_replace(substr, r'(%|_|\\)', r'\\\1', 'g')
    return self.like(sql.func.concat(escaped_substr, '%'))


@pxt.udf(is_method=True)
def strip(self: str, chars: Optional[str] = None) -> str:
    """
    Return a copy of string with leading and trailing characters removed.

    Equivalent to [`str.strip()`](https://docs.python.org/3/library/stdtypes.html#str.strip).

    Args:
        chars: The set of characters to be removed. If omitted or `None`, whitespace characters are removed.
    """
    return self.strip(chars)


@strip.to_sql
def _(self: sql.ColumnElement, chars: Optional[sql.ColumnElement] = None) -> sql.ColumnElement:
    return sql.func.trim(self, chars if chars is not None else whitespace)


@pxt.udf(is_method=True)
def swapcase(self: str) -> str:
    """
    Return a copy of string with uppercase characters converted to lowercase and vice versa.

    Equivalent to [`str.swapcase()`](https://docs.python.org/3/library/stdtypes.html#str.swapcase).
    """
    return self.swapcase()


@pxt.udf(is_method=True)
def title(self: str) -> str:
    """
    Return a titlecased version of string, i.e. words start with uppercase characters, all remaining cased characters
    are lowercase.

    Equivalent to [`str.title()`](https://docs.python.org/3/library/stdtypes.html#str.title).
    """
    return self.title()


@pxt.udf(is_method=True)
def upper(self: str) -> str:
    """
    Return a copy of string converted to uppercase.

    Equivalent to [`str.upper()`](https://docs.python.org/3/library/stdtypes.html#str.upper).
    """
    return self.upper()


@upper.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.func.upper(self)


@pxt.udf(is_method=True)
def wrap(self: str, width: int, **kwargs: Any) -> list[str]:
    """
    Wraps the single paragraph in string so every line is at most `width` characters long.
    Returns a list of output lines, without final newlines.

    Equivalent to [`textwrap.fill()`](https://docs.python.org/3/library/textwrap.html#textwrap.fill).

    Args:
        width: Maximum line width.
        kwargs: Additional keyword arguments to pass to `textwrap.fill()`.
    """
    return textwrap.wrap(self, width, **kwargs)


@pxt.udf(is_method=True)
def zfill(self: str, width: int) -> str:
    """
    Pad a numeric string with ASCII `0` on the left to a total length of `width`.

    Equivalent to [`str.zfill()`](https://docs.python.org/3/library/stdtypes.html#str.zfill).

    Args:
        width: Minimum width of resulting string.
    """
    return self.zfill(width)


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



import logging
from typing import Optional
from uuid import UUID

import sqlalchemy as sql

from pixeltable.metadata import register_converter
from pixeltable.metadata.converters.util import convert_table_md

_logger = logging.getLogger('pixeltable')


@register_converter(version=39)
def _(engine: sql.engine.Engine) -> None:
    convert_table_md(engine, table_modifier=__table_modifier)


def __table_modifier(conn: sql.Connection, tbl_id: UUID, orig_table_md: dict, updated_table_md: dict) -> None:
    store_prefix = 'view' if orig_table_md['view_md'] is not None else 'tbl'
    store_name = f'{store_prefix}_{tbl_id.hex}'

    # Get the list of column names that need to be migrated
    col_names = find_error_columns(conn=conn, store_name=store_name)
    if len(col_names) == 0:
        _logger.info(f'No error columns found in table {store_name}. Skipping migration.')
        return

    # Check if the table exists, outside of the metadata we were given
    # There seem to be cases where the metadata is present in the catalog,
    # but the table itself is not in the database.
    check_table_sql = sql.text(f"""
        SELECT EXISTS (
            SELECT 1
            FROM information_schema.tables
            WHERE table_name = '{store_name}'
        )
    """)
    table_exists = conn.execute(check_table_sql).scalar()
    if not table_exists:
        _logger.warning(f'Table {store_name} does not exist. Skipping migration.')
        return

    return migrate_error_to_cellmd_columns(conn, store_name, col_names)


def find_error_columns(conn: sql.Connection, store_name: str) -> list[str]:
    """
    Return and errormsg or errortype columns in the given table

    Args:
        conn: SQLAlchemy connection
        store_name: Name of the table to check

    Returns:
        List of column name roots (root_errormsg, root_errortype)
    """
    check_columns_sql = sql.text(f"""
        SELECT column_name
        FROM information_schema.columns
        WHERE table_name = '{store_name}'
    """)
    found_columns = [
        row[0]
        for row in conn.execute(check_columns_sql)
        if row[0].endswith('_errormsg') or row[0].endswith('_errortype')
    ]
    column_roots = {s.removesuffix('_errormsg').removesuffix('_errortype') for s in found_columns}
    return [*column_roots]


def migrate_error_to_cellmd_columns(
    conn: sql.Connection, store_name: str, col_names: list[str], backup_table: Optional[str] = None
) -> None:
    """
    Safe version with error handling and optional backup.

    Args:
        engine: SQLAlchemy engine
        store_name: Name of the table to modify
        col_names: List of column name prefixes
        backup_table: Optional name for backup table

    Usage:
        migrate_error_to_cellmd_columns(engine, 'my_table', ['columnname'], 'my_table_backup')
    """

    try:
        # Optional: Create backup
        if backup_table:
            backup_sql = sql.text(f"""
                CREATE TABLE {backup_table} AS SELECT * FROM {store_name}
            """)
            conn.execute(backup_sql)
            _logger.info(f'Backup created: {backup_table}')

        # Step 1: Add new columns
        add_column_str = ', '.join(f'ADD COLUMN {col}_cellmd JSONB DEFAULT NULL' for col in col_names)
        add_column_sql = sql.text(f'ALTER TABLE {store_name} {add_column_str}')
        conn.execute(add_column_sql)
        _logger.info(f'Added columns: {", ".join(f"{col}_cellmd" for col in col_names)}')

        # Step 2: Populate new columns
        set_column_str = ', '.join(
            [
                f'{col}_cellmd = CASE WHEN {col}_errormsg IS NULL OR {col}_errortype IS NULL '
                f"THEN NULL ELSE jsonb_build_object('errormsg', {col}_errormsg, 'errortype', {col}_errortype) END"
                for col in col_names
            ]
        )
        populate_sql = sql.text(f'UPDATE {store_name} SET {set_column_str}')
        result = conn.execute(populate_sql)
        _logger.info(f'Updated {result.rowcount} rows')

        # Step 3: Drop old columns
        drop_columns_str = ', '.join(
            [f'DROP COLUMN IF EXISTS {col}_errormsg, DROP COLUMN IF EXISTS {col}_errortype' for col in col_names]
        )
        drop_columns_sql = sql.text(f'ALTER TABLE {store_name} {drop_columns_str}')
        conn.execute(drop_columns_sql)
        _logger.info(f'Dropped columns: {", ".join(f"{col}_errormsg, {col}_errortype" for col in col_names)}')
        _logger.info(f'Migration completed successfully for table: {store_name}')

    except sql.exc.SQLAlchemyError as e:
        _logger.error(f'Migration for table {store_name} failed: {e}')
        raise



from __future__ import annotations

import logging
import os
import shutil
from pathlib import Path
from typing import Any, ClassVar, Optional, TypeVar

import toml

from pixeltable import exceptions as excs

_logger = logging.getLogger('pixeltable')

T = TypeVar('T')


class Config:
    """
    The (global) Pixeltable configuration, as loaded from `config.toml`. Provides methods for retrieving
    configuration values, which can be set in the config file or as environment variables.
    """

    __instance: ClassVar[Optional[Config]] = None

    __home: Path
    __config_file: Path
    __config_overrides: dict[str, Any]
    __config_dict: dict[str, Any]

    def __init__(self, config_overrides: dict[str, Any]) -> None:
        assert self.__instance is None, 'Config is a singleton; use Config.get() to access the instance'

        for var in config_overrides:
            if var not in KNOWN_CONFIG_OVERRIDES:
                raise excs.Error(f'Unrecognized configuration variable: {var}')

        self.__config_overrides = config_overrides

        self.__home = Path(self.lookup_env('pixeltable', 'home', str(Path.home() / '.pixeltable')))
        if self.__home.exists() and not self.__home.is_dir():
            raise excs.Error(f'Not a directory: {self.__home}')
        if not self.__home.exists():
            print(f'Creating a Pixeltable instance at: {self.__home}')
            self.__home.mkdir()

        self.__config_file = Path(self.lookup_env('pixeltable', 'config', str(self.__home / 'config.toml')))

        self.__config_dict: dict[str, Any]
        if os.path.isfile(self.__config_file):
            with open(self.__config_file, 'r', encoding='utf-8') as stream:
                try:
                    self.__config_dict = toml.load(stream)
                except Exception as exc:
                    raise excs.Error(f'Could not read config file: {self.__config_file}') from exc
            for section, section_dict in self.__config_dict.items():
                if section not in KNOWN_CONFIG_OPTIONS:
                    raise excs.Error(f'Unrecognized section {section!r} in config file: {self.__config_file}')
                for key in section_dict:
                    if key not in KNOWN_CONFIG_OPTIONS[section]:
                        raise excs.Error(f"Unrecognized option '{section}.{key}' in config file: {self.__config_file}")
        else:
            self.__config_dict = self.__create_default_config(self.__config_file)
            with open(self.__config_file, 'w', encoding='utf-8') as stream:
                try:
                    toml.dump(self.__config_dict, stream)
                except Exception as exc:
                    raise excs.Error(f'Could not write config file: {self.__config_file}') from exc
            _logger.info(f'Created default config file at: {self.__config_file}')

    @property
    def home(self) -> Path:
        return self.__home

    @property
    def config_file(self) -> Path:
        return self.__config_file

    @classmethod
    def get(cls) -> Config:
        cls.init({})
        return cls.__instance

    @classmethod
    def init(cls, config_overrides: dict[str, Any]) -> None:
        if cls.__instance is None:
            cls.__instance = cls(config_overrides)
        elif len(config_overrides) > 0:
            raise excs.Error(
                'Pixeltable has already been initialized; cannot specify new config values in the same session'
            )

    @classmethod
    def __create_default_config(cls, config_path: Path) -> dict[str, Any]:
        free_disk_space_bytes = shutil.disk_usage(config_path.parent).free
        # Default cache size is 1/5 of free disk space
        file_cache_size_g = free_disk_space_bytes / 5 / (1 << 30)
        return {'pixeltable': {'file_cache_size_g': round(file_cache_size_g, 1), 'hide_warnings': False}}

    def lookup_env(self, section: str, key: str, default: Any = None) -> Any:
        override_var = f'{section}.{key}'
        env_var = f'{section.upper()}_{key.upper()}'
        if override_var in self.__config_overrides:
            return self.__config_overrides[override_var]
        if env_var in os.environ:
            return os.environ[env_var]
        return default

    def get_value(self, key: str, expected_type: type[T], section: str = 'pixeltable') -> Optional[T]:
        value = self.lookup_env(section, key)  # Try to get from environment first
        # Next try the config file
        if value is None and section in self.__config_dict and key in self.__config_dict[section]:
            value = self.__config_dict[section][key]

        if value is None:
            return None  # Not specified

        try:
            if expected_type is bool and isinstance(value, str):
                if value.lower() not in ('true', 'false'):
                    raise excs.Error(f'Invalid value for configuration parameter {section}.{key}: {value}')
                return value.lower() == 'true'  # type: ignore[return-value]
            return expected_type(value)  # type: ignore[call-arg]
        except (ValueError, TypeError) as exc:
            raise excs.Error(f'Invalid value for configuration parameter {section}.{key}: {value}') from exc

    def get_string_value(self, key: str, section: str = 'pixeltable') -> Optional[str]:
        return self.get_value(key, str, section)

    def get_int_value(self, key: str, section: str = 'pixeltable') -> Optional[int]:
        return self.get_value(key, int, section)

    def get_float_value(self, key: str, section: str = 'pixeltable') -> Optional[float]:
        return self.get_value(key, float, section)

    def get_bool_value(self, key: str, section: str = 'pixeltable') -> Optional[bool]:
        return self.get_value(key, bool, section)


KNOWN_CONFIG_OPTIONS = {
    'pixeltable': {
        'home': 'Path to the Pixeltable home directory',
        'config': 'Path to the Pixeltable config file',
        'pgdata': 'Path to the Pixeltable postgres data directory',
        'db': 'Postgres database name',
        'file_cache_size_g': 'Size of the file cache in GB',
        'time_zone': 'Default time zone for timestamps',
        'hide_warnings': 'Hide warnings from the console',
        'verbosity': 'Verbosity level for console output',
        'api_key': 'API key for Pixeltable cloud',
    },
    'anthropic': {'api_key': 'Anthropic API key'},
    'bedrock': {'api_key': 'AWS Bedrock API key'},
    'deepseek': {'api_key': 'Deepseek API key'},
    'fireworks': {'api_key': 'Fireworks API key'},
    'gemini': {'api_key': 'Gemini API key'},
    'groq': {'api_key': 'Groq API key'},
    'label_studio': {'api_key': 'Label Studio API key', 'url': 'Label Studio server URL'},
    'mistral': {'api_key': 'Mistral API key'},
    'openai': {'api_key': 'OpenAI API key'},
    'replicate': {'api_token': 'Replicate API token'},
    'together': {'api_key': 'Together API key'},
    'pypi': {'api_key': 'PyPI API key (for internal use only)'},
}


KNOWN_CONFIG_OVERRIDES = {
    f'{section}.{key}': info
    for section, section_dict in KNOWN_CONFIG_OPTIONS.items()
    for key, info in section_dict.items()
}



from __future__ import annotations

import glob
import hashlib
import logging
import os
import warnings
from collections import OrderedDict, defaultdict
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import NamedTuple, Optional
from uuid import UUID

import pixeltable.exceptions as excs
from pixeltable.config import Config
from pixeltable.env import Env

_logger = logging.getLogger('pixeltable')


@dataclass
class CacheEntry:
    key: str
    tbl_id: UUID
    col_id: int
    size: int
    last_used: datetime
    ext: str

    @property
    def path(self) -> Path:
        return Env.get().file_cache_dir / f'{self.tbl_id.hex}_{self.col_id}_{self.key}{self.ext}'

    @classmethod
    def from_file(cls, path: Path) -> CacheEntry:
        components = path.stem.split('_')
        assert len(components) == 3
        tbl_id = UUID(components[0])
        col_id = int(components[1])
        key = components[2]
        file_info = os.stat(str(path))
        # We use the last modified time (file_info.st_mtime) as the timestamp; `FileCache` will touch the file
        # each time it is retrieved, so that the mtime of the file will always represent the last used time of
        # the cache entry.
        last_used = datetime.fromtimestamp(file_info.st_mtime, tz=timezone.utc)
        return cls(key, tbl_id, col_id, file_info.st_size, last_used, path.suffix)


class FileCache:
    """
    A local cache of external (eg, S3) file references in cells of a stored table (ie, table or view).

    Cache entries are identified by a hash of the file url and stored in Env.filecache_dir. The time of last
    access of a cache entries is its file's mtime.

    TODO:
    - implement MRU eviction for queries that exceed the capacity
    """

    __instance: Optional[FileCache] = None

    cache: OrderedDict[str, CacheEntry]
    total_size: int
    capacity_bytes: int
    num_requests: int
    num_hits: int
    num_evictions: int
    keys_retrieved: set[str]  # keys retrieved (downloaded or accessed) this session
    keys_evicted_after_retrieval: set[str]  # keys that were evicted after having been retrieved this session

    # A key is added to this set when it is already present in `keys_evicted_this_session` and is downloaded again.
    # In other words, for a key to be added to this set, the following sequence of events must occur in this order:
    # - It is retrieved during this session (either because it was newly downloaded, or because it was in the cache
    #   at the start of the session and was accessed at some point during the session)
    # - It is subsequently evicted
    # - It is subsequently retrieved a second time ("download after a previous retrieval")
    # The contents of this set will be used to generate a more informative warning.
    evicted_working_set_keys: set[str]
    new_redownload_witnessed: bool  # whether a new re-download has occurred since the last time a warning was issued

    class FileCacheColumnStats(NamedTuple):
        tbl_id: UUID
        col_id: int
        num_files: int
        total_size: int

    class FileCacheStats(NamedTuple):
        total_size: int
        num_requests: int
        num_hits: int
        num_evictions: int
        column_stats: list[FileCache.FileCacheColumnStats]

    @classmethod
    def get(cls) -> FileCache:
        if cls.__instance is None:
            cls.init()
        return cls.__instance

    @classmethod
    def init(cls) -> None:
        cls.__instance = cls()

    def __init__(self) -> None:
        self.cache = OrderedDict()
        self.total_size = 0
        self.capacity_bytes = int(Env.get()._file_cache_size_g * (1 << 30))
        self.num_requests = 0
        self.num_hits = 0
        self.num_evictions = 0
        self.keys_retrieved = set()
        self.keys_evicted_after_retrieval = set()
        self.evicted_working_set_keys = set()
        self.new_redownload_witnessed = False
        paths = glob.glob(str(Env.get().file_cache_dir / '*'))
        entries = [CacheEntry.from_file(Path(path_str)) for path_str in paths]
        # we need to insert entries in access order
        entries.sort(key=lambda e: e.last_used)
        for entry in entries:
            self.cache[entry.key] = entry
            self.total_size += entry.size

    def avg_file_size(self) -> int:
        if len(self.cache) == 0:
            return 0
        return int(self.total_size / len(self.cache))

    def num_files(self, tbl_id: Optional[UUID] = None) -> int:
        if tbl_id is None:
            return len(self.cache)
        return sum(e.tbl_id == tbl_id for e in self.cache.values())

    def clear(self, tbl_id: Optional[UUID] = None) -> None:
        """
        For testing purposes: allow resetting capacity and stats.
        """
        if tbl_id is None:
            # We need to store the entries to remove in a list, because we can't remove items from a dict
            # while iterating
            entries_to_remove = list(self.cache.values())
            _logger.debug(f'clearing {self.num_files()} entries from file cache')
            self.num_requests, self.num_hits, self.num_evictions = 0, 0, 0
            self.keys_retrieved.clear()
            self.keys_evicted_after_retrieval.clear()
            self.new_redownload_witnessed = False
        else:
            entries_to_remove = [e for e in self.cache.values() if e.tbl_id == tbl_id]
            _logger.debug(f'clearing {self.num_files(tbl_id)} entries from file cache for table {tbl_id}')
        for entry in entries_to_remove:
            os.remove(entry.path)
            del self.cache[entry.key]
            self.total_size -= entry.size

    def emit_eviction_warnings(self) -> None:
        if self.new_redownload_witnessed:
            # Compute the additional capacity that would be needed in order to retain all the re-downloaded files
            extra_capacity_needed = sum(self.cache[key].size for key in self.evicted_working_set_keys)
            suggested_cache_size = self.capacity_bytes + extra_capacity_needed + (1 << 30)
            warnings.warn(
                f'{len(self.evicted_working_set_keys)} media file(s) had to be downloaded multiple times this session, '
                'because they were evicted\nfrom the file cache after their first access. The total size '
                f'of the evicted file(s) is {round(extra_capacity_needed / (1 << 30), 1)} GiB.\n'
                f'Consider increasing the cache size to at least {round(suggested_cache_size / (1 << 30), 1)} GiB '
                f'(it is currently {round(self.capacity_bytes / (1 << 30), 1)} GiB).\n'
                f'You can do this by setting the value of `file_cache_size_g` in: {Config.get().config_file}',
                excs.PixeltableWarning,
                stacklevel=2,
            )
            self.new_redownload_witnessed = False

    def _url_hash(self, url: str) -> str:
        h = hashlib.sha256()
        h.update(url.encode())
        return h.hexdigest()

    def lookup(self, url: str) -> Optional[Path]:
        self.num_requests += 1
        key = self._url_hash(url)
        entry = self.cache.get(key, None)
        if entry is None:
            _logger.debug(f'file cache miss for {url}')
            return None
        # update mtime and cache
        path = entry.path
        path.touch(exist_ok=True)
        file_info = os.stat(str(path))
        entry.last_used = datetime.fromtimestamp(file_info.st_mtime)
        self.cache.move_to_end(key, last=True)
        self.num_hits += 1
        self.keys_retrieved.add(key)
        _logger.debug(f'file cache hit for {url}')
        return path

    def add(self, tbl_id: UUID, col_id: int, url: str, path: Path) -> Path:
        """Adds url at 'path' to cache and returns its new path.
        'path' will not be accessible after this call. Retains the extension of 'path'.
        """
        file_info = os.stat(str(path))
        self.ensure_capacity(file_info.st_size)
        key = self._url_hash(url)
        assert key not in self.cache
        if key in self.keys_evicted_after_retrieval:
            # This key was evicted after being retrieved earlier this session, and is now being retrieved again.
            # Add it to `keys_multiply_downloaded` so that we may generate a warning later.
            self.evicted_working_set_keys.add(key)
            self.new_redownload_witnessed = True
        self.keys_retrieved.add(key)
        entry = CacheEntry(
            key, tbl_id, col_id, file_info.st_size, datetime.fromtimestamp(file_info.st_mtime), path.suffix
        )
        self.cache[key] = entry
        self.total_size += entry.size
        new_path = entry.path
        os.rename(str(path), str(new_path))
        new_path.touch(exist_ok=True)
        _logger.debug(f'added entry for cell {url} to file cache')
        return new_path

    def ensure_capacity(self, size: int) -> None:
        """
        Evict entries from the cache until there is at least 'size' bytes of free space.
        """
        while len(self.cache) > 0 and self.total_size + size > self.capacity_bytes:
            _, lru_entry = self.cache.popitem(last=False)
            self.total_size -= lru_entry.size
            self.num_evictions += 1
            if lru_entry.key in self.keys_retrieved:
                # This key was retrieved at some point earlier this session and is now being evicted.
                # Make a record of the eviction, so that we can generate a warning later if the key is retrieved again.
                self.keys_evicted_after_retrieval.add(lru_entry.key)
            os.remove(str(lru_entry.path))
            _logger.debug(
                f'evicted entry for cell {lru_entry.key} from file cache (of size {lru_entry.size // (1 << 20)} MiB)'
            )

    def set_capacity(self, capacity_bytes: int) -> None:
        self.capacity_bytes = capacity_bytes
        self.ensure_capacity(0)  # evict entries if necessary

    def stats(self) -> FileCacheStats:
        # collect column stats
        # (tbl_id, col_id) -> (num_files, total_size)
        d: dict[tuple[UUID, int], list[int]] = defaultdict(lambda: [0, 0])
        for entry in self.cache.values():
            t = d[entry.tbl_id, entry.col_id]
            t[0] += 1
            t[1] += entry.size
        col_stats = [
            self.FileCacheColumnStats(tbl_id, col_id, num_files, size)
            for (tbl_id, col_id), (num_files, size) in d.items()
        ]
        col_stats.sort(key=lambda e: e[3], reverse=True)
        return self.FileCacheStats(self.total_size, self.num_requests, self.num_hits, self.num_evictions, col_stats)

    def debug_print(self) -> None:
        for entry in self.cache.values():
            _logger.debug(f'CacheEntry: tbl_id={entry.tbl_id}, col_id={entry.col_id}, size={entry.size}')



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

from pixeltable.type_system import ColumnType


class ComponentIterator(ABC):
    """Base class for Pixeltable iterators."""

    @classmethod
    @abstractmethod
    def input_schema(cls) -> dict[str, ColumnType]:
        """Provide the Pixeltable types of the init() parameters

        The keys need to match the names of the init() parameters. This is equivalent to the parameters_types
        parameter of the @function decorator.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def output_schema(cls, *args: Any, **kwargs: Any) -> tuple[dict[str, ColumnType], list[str]]:
        """Specify the dictionary returned by next() and a list of unstored column names

        Returns:
            a dictionary which is turned into a list of columns in the output table
            a list of unstored column names
        """
        raise NotImplementedError

    def __iter__(self) -> ComponentIterator:
        return self

    @abstractmethod
    def __next__(self) -> dict[str, Any]:
        """Return the next element of the iterator as a dictionary or raise StopIteration"""
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None:
        """Close the iterator and release all resources"""
        raise NotImplementedError

    @abstractmethod
    def set_pos(self, pos: int) -> None:
        """Set the iterator position to pos"""
        raise NotImplementedError

    @classmethod
    def create(cls, **kwargs: Any) -> tuple[type[ComponentIterator], dict[str, Any]]:
        return cls, kwargs



from __future__ import annotations

import datetime
from typing import Any, Optional

import numpy as np
import sqlalchemy as sql

import pixeltable.type_system as ts
from pixeltable.env import Env

from .data_row import DataRow
from .expr import Expr
from .row_builder import RowBuilder
from .sql_element_cache import SqlElementCache


class Literal(Expr):
    val: Any

    def __init__(self, val: Any, col_type: Optional[ts.ColumnType] = None):
        if col_type is not None:
            val = col_type.create_literal(val)
        else:
            # try to determine a type for val
            col_type = ts.ColumnType.infer_literal_type(val)
            if col_type is None:
                raise TypeError(f'Not a valid literal: {val}')
        super().__init__(col_type)
        if isinstance(val, datetime.datetime):
            # Normalize the datetime to UTC: all timestamps are stored as UTC (both in the database and in literals)
            if val.tzinfo is None:
                # We have a naive datetime. Modify it to use the configured default time zone
                default_tz = Env.get().default_time_zone
                if default_tz is not None:
                    val = val.replace(tzinfo=default_tz)
            # Now convert to UTC
            val = val.astimezone(datetime.timezone.utc)
        if isinstance(val, tuple):
            # Tuples are stored as a list
            val = list(val)
        self.val = val
        self.id = self._create_id()

    def default_column_name(self) -> Optional[str]:
        return 'Literal'

    def __str__(self) -> str:
        if self.col_type.is_string_type():
            return f"'{self.val}'"
        if self.col_type.is_timestamp_type():
            assert isinstance(self.val, datetime.datetime)
            default_tz = Env.get().default_time_zone
            return f"'{self.val.astimezone(default_tz).isoformat()}'"
        if self.col_type.is_date_type():
            assert isinstance(self.val, datetime.date)
            return f"'{self.val.isoformat()}'"
        if self.col_type.is_array_type():
            assert isinstance(self.val, np.ndarray)
            return str(self.val.tolist())
        return str(self.val)

    def __repr__(self) -> str:
        return f'Literal({self.val!r})'

    def _equals(self, other: Literal) -> bool:
        return self.val == other.val

    def _id_attrs(self) -> list[tuple[str, Any]]:
        return [*super()._id_attrs(), ('val', self.val)]

    def sql_expr(self, _: SqlElementCache) -> Optional[sql.ColumnElement]:
        # Return a sql object so that constants can participate in SQL expressions
        return sql.sql.expression.literal(self.val, type_=self.col_type.to_sa_type())

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        # this will be called, even though sql_expr() does not return None
        data_row[self.slot_idx] = self.val

    def _as_dict(self) -> dict:
        # For some types, we need to explicitly record their type, because JSON does not know
        # how to interpret them unambiguously
        if self.col_type.is_timestamp_type():
            assert isinstance(self.val, datetime.datetime)
            assert self.val.tzinfo == datetime.timezone.utc  # Must be UTC in a literal
            # Convert to ISO format in UTC (in keeping with the principle: all timestamps are
            # stored as UTC in the database)
            encoded_val = self.val.isoformat()
            return {'val': encoded_val, 'val_t': self.col_type._type.name, **super()._as_dict()}
        elif self.col_type.is_date_type():
            assert isinstance(self.val, datetime.date)
            encoded_val = self.val.isoformat()
            return {'val': encoded_val, 'val_t': self.col_type._type.name, **super()._as_dict()}
        elif self.col_type.is_array_type():
            assert isinstance(self.val, np.ndarray)
            return {'val': self.val.tolist(), 'val_t': self.col_type._type.name, **super()._as_dict()}
        else:
            return {'val': self.val, **super()._as_dict()}

    def as_literal(self) -> Optional[Literal]:
        return self

    @classmethod
    def _from_dict(cls, d: dict, components: list[Expr]) -> Literal:
        assert 'val' in d
        if 'val_t' in d:
            val_t = d['val_t']
            if val_t == ts.ColumnType.Type.DATE.name:
                dt = datetime.date.fromisoformat(d['val'])
                return cls(dt)
            elif val_t == ts.ColumnType.Type.TIMESTAMP.name:
                dt = datetime.datetime.fromisoformat(d['val'])
                assert dt.tzinfo == datetime.timezone.utc  # Must be UTC in the database
                return cls(dt)
            elif val_t == ts.ColumnType.Type.ARRAY.name:
                arrays = np.array(d['val'])
                return cls(arrays)
        return cls(d['val'])



from typing import Any, Optional

import sqlalchemy as sql

from pixeltable.metadata import register_converter
from pixeltable.metadata.converters.util import convert_table_md


@register_converter(version=26)
def _(engine: sql.engine.Engine) -> None:
    convert_table_md(engine, substitution_fn=__substitute_md)


def __substitute_md(k: Optional[str], v: Any) -> Optional[tuple[Optional[str], Any]]:
    import pixeltable.type_system as ts
    from pixeltable.exprs.literal import Literal

    if k == 'limit_val' and v is not None:
        assert isinstance(v, int)
        newv = Literal(v, ts.IntType(nullable=False)).as_dict()
        return k, newv

    return None



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs)
that wrap various endpoints from the OpenAI API. In order to use them, you must
first `pip install openai` and configure your OpenAI credentials, as described in
the [Working with OpenAI](https://pixeltable.readme.io/docs/working-with-openai) tutorial.
"""

import base64
import datetime
import io
import json
import logging
import math
import pathlib
import re
from typing import TYPE_CHECKING, Any, Callable, Optional, Type

import httpx
import numpy as np
import PIL

import pixeltable as pxt
from pixeltable import env, exprs, type_system as ts
from pixeltable.func import Batch, Tools
from pixeltable.utils.code import local_public_names

if TYPE_CHECKING:
    import openai

_logger = logging.getLogger('pixeltable')


@env.register_client('openai')
def _(api_key: str) -> 'openai.AsyncOpenAI':
    import openai

    return openai.AsyncOpenAI(
        api_key=api_key,
        # recommended to increase limits for async client to avoid connection errors
        http_client=httpx.AsyncClient(limits=httpx.Limits(max_keepalive_connections=100, max_connections=500)),
    )


def _openai_client() -> 'openai.AsyncOpenAI':
    return env.Env.get().get_client('openai')


# models that share rate limits; see https://platform.openai.com/settings/organization/limits for details
_shared_rate_limits = {
    'gpt-4-turbo': [
        'gpt-4-turbo',
        'gpt-4-turbo-latest',
        'gpt-4-turbo-2024-04-09',
        'gpt-4-turbo-preview',
        'gpt-4-0125-preview',
        'gpt-4-1106-preview',
    ],
    'gpt-4o': [
        'gpt-4o',
        'gpt-4o-latest',
        'gpt-4o-2024-05-13',
        'gpt-4o-2024-08-06',
        'gpt-4o-2024-11-20',
        'gpt-4o-audio-preview',
        'gpt-4o-audio-preview-2024-10-01',
        'gpt-4o-audio-preview-2024-12-17',
    ],
    'gpt-4o-mini': [
        'gpt-4o-mini',
        'gpt-4o-mini-latest',
        'gpt-4o-mini-2024-07-18',
        'gpt-4o-mini-audio-preview',
        'gpt-4o-mini-audio-preview-2024-12-17',
    ],
    'gpt-4o-mini-realtime-preview': [
        'gpt-4o-mini-realtime-preview',
        'gpt-4o-mini-realtime-preview-latest',
        'gpt-4o-mini-realtime-preview-2024-12-17',
    ],
}


def _rate_limits_pool(model: str) -> str:
    for model_family, models in _shared_rate_limits.items():
        if model in models:
            return f'rate-limits:openai:{model_family}'
    return f'rate-limits:openai:{model}'


class OpenAIRateLimitsInfo(env.RateLimitsInfo):
    retryable_errors: tuple[Type[Exception], ...]

    def __init__(self, get_request_resources: Callable[..., dict[str, int]]):
        super().__init__(get_request_resources)
        import openai

        self.retryable_errors = (
            # ConnectionError: we occasionally see this error when the AsyncConnectionPool is trying to close
            # expired connections
            # (AsyncConnectionPool._close_expired_connections() fails with ConnectionError when executing
            # 'await connection.aclose()', which is very likely a bug in AsyncConnectionPool)
            openai.APIConnectionError,
            # the following errors are retryable according to OpenAI's API documentation
            openai.RateLimitError,
            openai.APITimeoutError,
            openai.UnprocessableEntityError,
            openai.InternalServerError,
        )

    def get_retry_delay(self, exc: Exception) -> Optional[float]:
        import openai

        if not isinstance(exc, self.retryable_errors):
            return None
        assert isinstance(exc, openai.APIError)
        return 1.0


# RE pattern for duration in '*-reset' headers;
# examples: 1d2h3ms, 4m5.6s; # fractional seconds can be reported as 0.5s or 500ms
_header_duration_pattern = re.compile(r'(?:(\d+)d)?(?:(\d+)h)?(?:(\d+)ms)|(?:(\d+)m)?(?:([\d.]+)s)?')


def _parse_header_duration(duration_str: str) -> datetime.timedelta:
    match = _header_duration_pattern.match(duration_str)
    if not match:
        raise ValueError('Invalid duration format')

    days = int(match.group(1) or 0)
    hours = int(match.group(2) or 0)
    milliseconds = int(match.group(3) or 0)
    minutes = int(match.group(4) or 0)
    seconds = float(match.group(5) or 0)

    return datetime.timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds, milliseconds=milliseconds)


def _get_header_info(
    headers: httpx.Headers, *, requests: bool = True, tokens: bool = True
) -> tuple[Optional[tuple[int, int, datetime.datetime]], Optional[tuple[int, int, datetime.datetime]]]:
    assert requests or tokens
    now = datetime.datetime.now(tz=datetime.timezone.utc)

    requests_info: Optional[tuple[int, int, datetime.datetime]] = None
    if requests:
        requests_limit_str = headers.get('x-ratelimit-limit-requests')
        requests_limit = int(requests_limit_str) if requests_limit_str is not None else None
        requests_remaining_str = headers.get('x-ratelimit-remaining-requests')
        requests_remaining = int(requests_remaining_str) if requests_remaining_str is not None else None
        requests_reset_str = headers.get('x-ratelimit-reset-requests')
        requests_reset_ts = now + _parse_header_duration(requests_reset_str)
        requests_info = (requests_limit, requests_remaining, requests_reset_ts)

    tokens_info: Optional[tuple[int, int, datetime.datetime]] = None
    if tokens:
        tokens_limit_str = headers.get('x-ratelimit-limit-tokens')
        tokens_limit = int(tokens_limit_str) if tokens_limit_str is not None else None
        tokens_remaining_str = headers.get('x-ratelimit-remaining-tokens')
        tokens_remaining = int(tokens_remaining_str) if tokens_remaining_str is not None else None
        tokens_reset_str = headers.get('x-ratelimit-reset-tokens')
        tokens_reset_ts = now + _parse_header_duration(tokens_reset_str)
        tokens_info = (tokens_limit, tokens_remaining, tokens_reset_ts)

    return requests_info, tokens_info


#####################################
# Audio Endpoints


@pxt.udf
async def speech(input: str, *, model: str, voice: str, model_kwargs: Optional[dict[str, Any]] = None) -> pxt.Audio:
    """
    Generates audio from the input text.

    Equivalent to the OpenAI `audio/speech` API endpoint.
    For additional details, see: <https://platform.openai.com/docs/guides/text-to-speech>

    Request throttling:
    Applies the rate limit set in the config (section `openai.rate_limits`; use the model id as the key). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install openai`

    Args:
        input: The text to synthesize into speech.
        model: The model to use for speech synthesis.
        voice: The voice profile to use for speech synthesis. Supported options include:
            `alloy`, `echo`, `fable`, `onyx`, `nova`, and `shimmer`.
        model_kwargs: Additional keyword args for the OpenAI `audio/speech` API. For details on the available
            parameters, see: <https://platform.openai.com/docs/api-reference/audio/createSpeech>

    Returns:
        An audio file containing the synthesized speech.

    Examples:
        Add a computed column that applies the model `tts-1` to an existing Pixeltable column `tbl.text`
        of the table `tbl`:

        >>> tbl.add_computed_column(audio=speech(tbl.text, model='tts-1', voice='nova'))
    """
    if model_kwargs is None:
        model_kwargs = {}

    content = await _openai_client().audio.speech.create(input=input, model=model, voice=voice, **model_kwargs)
    ext = model_kwargs.get('response_format', 'mp3')
    output_filename = str(env.Env.get().create_tmp_path(f'.{ext}'))
    content.write_to_file(output_filename)
    return output_filename


@pxt.udf
async def transcriptions(audio: pxt.Audio, *, model: str, model_kwargs: Optional[dict[str, Any]] = None) -> dict:
    """
    Transcribes audio into the input language.

    Equivalent to the OpenAI `audio/transcriptions` API endpoint.
    For additional details, see: <https://platform.openai.com/docs/guides/speech-to-text>

    Request throttling:
    Applies the rate limit set in the config (section `openai.rate_limits`; use the model id as the key). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install openai`

    Args:
        audio: The audio to transcribe.
        model: The model to use for speech transcription.
        model_kwargs: Additional keyword args for the OpenAI `audio/transcriptions` API. For details on the available
            parameters, see: <https://platform.openai.com/docs/api-reference/audio/createTranscription>

    Returns:
        A dictionary containing the transcription and other metadata.

    Examples:
        Add a computed column that applies the model `whisper-1` to an existing Pixeltable column `tbl.audio`
        of the table `tbl`:

        >>> tbl.add_computed_column(transcription=transcriptions(tbl.audio, model='whisper-1', language='en'))
    """
    if model_kwargs is None:
        model_kwargs = {}

    file = pathlib.Path(audio)
    transcription = await _openai_client().audio.transcriptions.create(file=file, model=model, **model_kwargs)
    return transcription.dict()


@pxt.udf
async def translations(audio: pxt.Audio, *, model: str, model_kwargs: Optional[dict[str, Any]] = None) -> dict:
    """
    Translates audio into English.

    Equivalent to the OpenAI `audio/translations` API endpoint.
    For additional details, see: <https://platform.openai.com/docs/guides/speech-to-text>

    Request throttling:
    Applies the rate limit set in the config (section `openai.rate_limits`; use the model id as the key). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install openai`

    Args:
        audio: The audio to translate.
        model: The model to use for speech transcription and translation.
        model_kwargs: Additional keyword args for the OpenAI `audio/translations` API. For details on the available
            parameters, see: <https://platform.openai.com/docs/api-reference/audio/createTranslation>

    Returns:
        A dictionary containing the translation and other metadata.

    Examples:
        Add a computed column that applies the model `whisper-1` to an existing Pixeltable column `tbl.audio`
        of the table `tbl`:

        >>> tbl.add_computed_column(translation=translations(tbl.audio, model='whisper-1', language='en'))
    """
    if model_kwargs is None:
        model_kwargs = {}

    file = pathlib.Path(audio)
    translation = await _openai_client().audio.translations.create(file=file, model=model, **model_kwargs)
    return translation.dict()


#####################################
# Chat Endpoints


def _default_max_tokens(model: str) -> int:
    if (
        _is_model_family(model, 'gpt-4o-realtime')
        or _is_model_family(model, 'gpt-4o-mini-realtime')
        or _is_model_family(model, 'gpt-4-turbo')
        or _is_model_family(model, 'gpt-3.5-turbo')
    ):
        return 4096
    if _is_model_family(model, 'gpt-4'):
        return 8192  # All other gpt-4 models (will not match on gpt-4o models)
    if _is_model_family(model, 'gpt-4o') or _is_model_family(model, 'gpt-4.5-preview'):
        return 16384  # All other gpt-4o / gpt-4.5 models
    if _is_model_family(model, 'o1-preview'):
        return 32768
    if _is_model_family(model, 'o1-mini'):
        return 65536
    if _is_model_family(model, 'o1') or _is_model_family(model, 'o3'):
        return 100000  # All other o1 / o3 models
    return 100000  # global default


def _is_model_family(model: str, family: str) -> bool:
    # `model.startswith(family)` would be a simpler match, but increases the risk of false positives.
    # We use a slightly more complicated criterion to make things a little less error prone.
    return model == family or model.startswith(f'{family}-')


def _chat_completions_get_request_resources(
    messages: list, model: str, model_kwargs: Optional[dict[str, Any]]
) -> dict[str, int]:
    if model_kwargs is None:
        model_kwargs = {}

    max_completion_tokens = model_kwargs.get('max_completion_tokens')
    max_tokens = model_kwargs.get('max_tokens')
    n = model_kwargs.get('n')

    completion_tokens = (n or 1) * (max_completion_tokens or max_tokens or _default_max_tokens(model))

    num_tokens = 0.0
    for message in messages:
        num_tokens += 4  # every message follows <im_start>{role/name}\n{content}<im_end>\n
        for key, value in message.items():
            num_tokens += len(value) / 4
            if key == 'name':  # if there's a name, the role is omitted
                num_tokens -= 1  # role is always required and always 1 token
    num_tokens += 2  # every reply is primed with <im_start>assistant
    return {'requests': 1, 'tokens': int(num_tokens) + completion_tokens}


@pxt.udf
async def chat_completions(
    messages: list,
    *,
    model: str,
    model_kwargs: Optional[dict[str, Any]] = None,
    tools: Optional[list[dict[str, Any]]] = None,
    tool_choice: Optional[dict[str, Any]] = None,
) -> dict:
    """
    Creates a model response for the given chat conversation.

    Equivalent to the OpenAI `chat/completions` API endpoint.
    For additional details, see: <https://platform.openai.com/docs/guides/chat-completions>

    Request throttling:
    Uses the rate limit-related headers returned by the API to throttle requests adaptively, based on available
    request and token capacity. No configuration is necessary.

    __Requirements:__

    - `pip install openai`

    Args:
        messages: A list of messages to use for chat completion, as described in the OpenAI API documentation.
        model: The model to use for chat completion.
        model_kwargs: Additional keyword args for the OpenAI `chat/completions` API. For details on the available
            parameters, see: <https://platform.openai.com/docs/api-reference/chat/create>

    Returns:
        A dictionary containing the response and other metadata.

    Examples:
        Add a computed column that applies the model `gpt-4o-mini` to an existing Pixeltable column `tbl.prompt`
        of the table `tbl`:

        >>> messages = [
                {'role': 'system', 'content': 'You are a helpful assistant.'},
                {'role': 'user', 'content': tbl.prompt}
            ]
            tbl.add_computed_column(response=chat_completions(messages, model='gpt-4o-mini'))
    """
    if model_kwargs is None:
        model_kwargs = {}

    if tools is not None:
        model_kwargs['tools'] = [{'type': 'function', 'function': tool} for tool in tools]

    if tool_choice is not None:
        if tool_choice['auto']:
            model_kwargs['tool_choice'] = 'auto'
        elif tool_choice['required']:
            model_kwargs['tool_choice'] = 'required'
        else:
            assert tool_choice['tool'] is not None
            model_kwargs['tool_choice'] = {'type': 'function', 'function': {'name': tool_choice['tool']}}

    if tool_choice is not None and not tool_choice['parallel_tool_calls']:
        model_kwargs['parallel_tool_calls'] = False

    # make sure the pool info exists prior to making the request
    resource_pool = _rate_limits_pool(model)
    rate_limits_info = env.Env.get().get_resource_pool_info(
        resource_pool, lambda: OpenAIRateLimitsInfo(_chat_completions_get_request_resources)
    )

    result = await _openai_client().chat.completions.with_raw_response.create(
        messages=messages, model=model, **model_kwargs
    )

    requests_info, tokens_info = _get_header_info(result.headers)
    rate_limits_info.record(requests=requests_info, tokens=tokens_info)

    return json.loads(result.text)


def _vision_get_request_resources(
    prompt: str, image: PIL.Image.Image, model: str, model_kwargs: Optional[dict[str, Any]] = None
) -> dict[str, int]:
    if model_kwargs is None:
        model_kwargs = {}

    max_completion_tokens = model_kwargs.get('max_completion_tokens')
    max_tokens = model_kwargs.get('max_tokens')
    n = model_kwargs.get('n')

    completion_tokens = (n or 1) * (max_completion_tokens or max_tokens or _default_max_tokens(model))
    prompt_tokens = len(prompt) / 4

    # calculate image tokens based on
    # https://platform.openai.com/docs/guides/vision/calculating-costs#calculating-costs
    # assuming detail='high' (which appears to be the default, according to community forum posts)

    # number of 512x512 crops; ceil(): partial crops still count as full crops
    crops_width = math.ceil(image.width / 512)
    crops_height = math.ceil(image.height / 512)
    total_crops = crops_width * crops_height

    base_tokens = 85  # base cost for the initial 512x512 overview
    crop_tokens = 170  # cost per additional 512x512 crop
    img_tokens = base_tokens + (crop_tokens * total_crops)

    total_tokens = (
        prompt_tokens
        + img_tokens
        + completion_tokens
        + 4  # for <im_start>{role/name}\n{content}<im_end>\n
        + 2  # for reply's <im_start>assistant
    )
    return {'requests': 1, 'tokens': int(total_tokens)}


@pxt.udf
async def vision(
    prompt: str, image: PIL.Image.Image, *, model: str, model_kwargs: Optional[dict[str, Any]] = None
) -> str:
    """
    Analyzes an image with the OpenAI vision capability. This is a convenience function that takes an image and
    prompt, and constructs a chat completion request that utilizes OpenAI vision.

    For additional details, see: <https://platform.openai.com/docs/guides/vision>

    Request throttling:
    Uses the rate limit-related headers returned by the API to throttle requests adaptively, based on available
    request and token capacity. No configuration is necessary.

    __Requirements:__

    - `pip install openai`

    Args:
        prompt: A prompt for the OpenAI vision request.
        image: The image to analyze.
        model: The model to use for OpenAI vision.

    Returns:
        The response from the OpenAI vision API.

    Examples:
        Add a computed column that applies the model `gpt-4o-mini` to an existing Pixeltable column `tbl.image`
        of the table `tbl`:

        >>> tbl.add_computed_column(response=vision("What's in this image?", tbl.image, model='gpt-4o-mini'))
    """
    if model_kwargs is None:
        model_kwargs = {}

    # TODO(aaron-siegel): Decompose CPU/GPU ops into separate functions
    bytes_arr = io.BytesIO()
    image.save(bytes_arr, format='png')
    b64_bytes = base64.b64encode(bytes_arr.getvalue())
    b64_encoded_image = b64_bytes.decode('utf-8')
    messages = [
        {
            'role': 'user',
            'content': [
                {'type': 'text', 'text': prompt},
                {'type': 'image_url', 'image_url': {'url': f'data:image/png;base64,{b64_encoded_image}'}},
            ],
        }
    ]

    # make sure the pool info exists prior to making the request
    resource_pool = _rate_limits_pool(model)
    rate_limits_info = env.Env.get().get_resource_pool_info(
        resource_pool, lambda: OpenAIRateLimitsInfo(_vision_get_request_resources)
    )

    result = await _openai_client().chat.completions.with_raw_response.create(
        messages=messages,  # type: ignore
        model=model,
        **model_kwargs,
    )

    requests_info, tokens_info = _get_header_info(result.headers)
    rate_limits_info.record(requests=requests_info, tokens=tokens_info)

    result = json.loads(result.text)
    return result['choices'][0]['message']['content']


#####################################
# Embeddings Endpoints

_embedding_dimensions_cache: dict[str, int] = {
    'text-embedding-ada-002': 1536,
    'text-embedding-3-small': 1536,
    'text-embedding-3-large': 3072,
}


def _embeddings_get_request_resources(input: list[str]) -> dict[str, int]:
    input_len = sum(len(s) for s in input)
    return {'requests': 1, 'tokens': int(input_len / 4)}


@pxt.udf(batch_size=32)
async def embeddings(
    input: Batch[str], *, model: str, model_kwargs: Optional[dict[str, Any]] = None
) -> Batch[pxt.Array[(None,), pxt.Float]]:
    """
    Creates an embedding vector representing the input text.

    Equivalent to the OpenAI `embeddings` API endpoint.
    For additional details, see: <https://platform.openai.com/docs/guides/embeddings>

    Request throttling:
    Uses the rate limit-related headers returned by the API to throttle requests adaptively, based on available
    request and token capacity. No configuration is necessary.

    __Requirements:__

    - `pip install openai`

    Args:
        input: The text to embed.
        model: The model to use for the embedding.
        model_kwargs: Additional keyword args for the OpenAI `embeddings` API. For details on the available
            parameters, see: <https://platform.openai.com/docs/api-reference/embeddings>

    Returns:
        An array representing the application of the given embedding to `input`.

    Examples:
        Add a computed column that applies the model `text-embedding-3-small` to an existing
        Pixeltable column `tbl.text` of the table `tbl`:

        >>> tbl.add_computed_column(embed=embeddings(tbl.text, model='text-embedding-3-small'))

        Add an embedding index to an existing column `text`, using the model `text-embedding-3-small`:

        >>> tbl.add_embedding_index(embedding=embeddings.using(model='text-embedding-3-small'))
    """
    if model_kwargs is None:
        model_kwargs = {}

    _logger.debug(f'embeddings: batch_size={len(input)}')
    resource_pool = _rate_limits_pool(model)
    rate_limits_info = env.Env.get().get_resource_pool_info(
        resource_pool, lambda: OpenAIRateLimitsInfo(_embeddings_get_request_resources)
    )
    result = await _openai_client().embeddings.with_raw_response.create(
        input=input, model=model, encoding_format='float', **model_kwargs
    )
    requests_info, tokens_info = _get_header_info(result.headers)
    rate_limits_info.record(requests=requests_info, tokens=tokens_info)
    return [np.array(data['embedding'], dtype=np.float64) for data in json.loads(result.content)['data']]


@embeddings.conditional_return_type
def _(model: str, model_kwargs: Optional[dict[str, Any]] = None) -> ts.ArrayType:
    dimensions: Optional[int] = None
    if model_kwargs is not None:
        dimensions = model_kwargs.get('dimensions')
    if dimensions is None:
        if model not in _embedding_dimensions_cache:
            # TODO: find some other way to retrieve a sample
            return ts.ArrayType((None,), dtype=ts.FloatType(), nullable=False)
        dimensions = _embedding_dimensions_cache.get(model)
    return ts.ArrayType((dimensions,), dtype=ts.FloatType(), nullable=False)


#####################################
# Images Endpoints


@pxt.udf
async def image_generations(
    prompt: str, *, model: str = 'dall-e-2', model_kwargs: Optional[dict[str, Any]] = None
) -> PIL.Image.Image:
    """
    Creates an image given a prompt.

    Equivalent to the OpenAI `images/generations` API endpoint.
    For additional details, see: <https://platform.openai.com/docs/guides/images>

    Request throttling:
    Applies the rate limit set in the config (section `openai.rate_limits`; use the model id as the key). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install openai`

    Args:
        prompt: Prompt for the image.
        model: The model to use for the generations.
        model_kwargs: Additional keyword args for the OpenAI `images/generations` API. For details on the available
            parameters, see: <https://platform.openai.com/docs/api-reference/images/create>

    Returns:
        The generated image.

    Examples:
        Add a computed column that applies the model `dall-e-2` to an existing
        Pixeltable column `tbl.text` of the table `tbl`:

        >>> tbl.add_computed_column(gen_image=image_generations(tbl.text, model='dall-e-2'))
    """
    if model_kwargs is None:
        model_kwargs = {}

    # TODO(aaron-siegel): Decompose CPU/GPU ops into separate functions
    result = await _openai_client().images.generate(
        prompt=prompt, model=model, response_format='b64_json', **model_kwargs
    )
    b64_str = result.data[0].b64_json
    b64_bytes = base64.b64decode(b64_str)
    img = PIL.Image.open(io.BytesIO(b64_bytes))
    img.load()
    return img


@image_generations.conditional_return_type
def _(model_kwargs: Optional[dict[str, Any]] = None) -> ts.ImageType:
    if model_kwargs is None or 'size' not in model_kwargs:
        # default size is 1024x1024
        return ts.ImageType(size=(1024, 1024))
    size = model_kwargs['size']
    x_pos = size.find('x')
    if x_pos == -1:
        return ts.ImageType()
    try:
        width, height = int(size[:x_pos]), int(size[x_pos + 1 :])
    except ValueError:
        return ts.ImageType()
    return ts.ImageType(size=(width, height))


#####################################
# Moderations Endpoints


@pxt.udf
async def moderations(input: str, *, model: str = 'omni-moderation-latest') -> dict:
    """
    Classifies if text is potentially harmful.

    Equivalent to the OpenAI `moderations` API endpoint.
    For additional details, see: <https://platform.openai.com/docs/guides/moderation>

    Request throttling:
    Applies the rate limit set in the config (section `openai.rate_limits`; use the model id as the key). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install openai`

    Args:
        input: Text to analyze with the moderations model.
        model: The model to use for moderations.

    For details on the other parameters, see: <https://platform.openai.com/docs/api-reference/moderations>

    Returns:
        Details of the moderations results.

    Examples:
        Add a computed column that applies the model `text-moderation-stable` to an existing
        Pixeltable column `tbl.input` of the table `tbl`:

        >>> tbl.add_computed_column(moderations=moderations(tbl.text, model='text-moderation-stable'))
    """
    result = await _openai_client().moderations.create(input=input, model=model)
    return result.dict()


@speech.resource_pool
@transcriptions.resource_pool
@translations.resource_pool
@image_generations.resource_pool
@moderations.resource_pool
def _(model: str) -> str:
    return f'request-rate:openai:{model}'


@chat_completions.resource_pool
@vision.resource_pool
@embeddings.resource_pool
def _(model: str) -> str:
    return _rate_limits_pool(model)


def invoke_tools(tools: Tools, response: exprs.Expr) -> exprs.InlineDict:
    """Converts an OpenAI response dict to Pixeltable tool invocation format and calls `tools._invoke()`."""
    return tools._invoke(_openai_response_to_pxt_tool_calls(response))


@pxt.udf
def _openai_response_to_pxt_tool_calls(response: dict) -> Optional[dict]:
    if 'tool_calls' not in response['choices'][0]['message'] or response['choices'][0]['message']['tool_calls'] is None:
        return None
    openai_tool_calls = response['choices'][0]['message']['tool_calls']
    pxt_tool_calls: dict[str, list[dict[str, Any]]] = {}
    for tool_call in openai_tool_calls:
        tool_name = tool_call['function']['name']
        if tool_name not in pxt_tool_calls:
            pxt_tool_calls[tool_name] = []
        pxt_tool_calls[tool_name].append({'args': json.loads(tool_call['function']['arguments'])})
    return pxt_tool_calls


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



from pathlib import Path
from typing import Union

from pyiceberg.catalog.sql import SqlCatalog


def sqlite_catalog(warehouse_path: Union[str, Path], name: str = 'pixeltable') -> SqlCatalog:
    """
    Instantiate a sqlite Iceberg catalog at the specified path. If no catalog exists, one will be created.
    """
    if isinstance(warehouse_path, str):
        warehouse_path = Path(warehouse_path)
    warehouse_path.mkdir(exist_ok=True)
    return SqlCatalog(name, uri=f'sqlite:///{warehouse_path}/catalog.db', warehouse=f'file://{warehouse_path}')



import datetime
import io
import json
from pathlib import Path
from typing import Any, Iterator, Sequence

import numpy as np
import PIL.Image
import torch
import torch.utils.data
import torchvision  # type: ignore[import-untyped]
from pyarrow import parquet

from pixeltable.type_system import ColumnType


class PixeltablePytorchDataset(torch.utils.data.IterableDataset):
    """
    PyTorch dataset interface for pixeltable data.
    NB. This class must inherit from torch.utils.data.IterableDataset for it
    to work with torch.utils.data.DataLoader.
    """

    def __init__(self, path: Path, image_format: str):
        """
        Args:
            path: path to directory containing parquet files
            image_format: 'np' or 'pt'. 'np' is RGB uint8 array,
                        'pt' is result of torchvision.transforms.ToTensor()
        """
        super().__init__()

        self.path = path
        self.image_format = image_format
        assert image_format in ('np', 'pt')
        column_type_path = path / '.pixeltable.column_types.json'
        assert column_type_path.exists(), f'missing {column_type_path}'
        with column_type_path.open() as f:
            column_types = json.load(f)
        self.column_types = {k: ColumnType.from_dict(v) for k, v in column_types.items()}
        self.part_metadata: list = parquet.ParquetDataset(str(path)).files

    def _unmarshall(self, k: str, v: Any) -> Any:
        if self.column_types[k].is_image_type():
            assert isinstance(v, bytes)
            im = PIL.Image.open(io.BytesIO(v))
            arr = np.array(im)  # will copy data to guarantee "WRITEABLE" flag assertion below.
            assert arr.flags['WRITEABLE']

            if self.image_format == 'np':
                return arr

            assert self.image_format == 'pt'

            # use arr instead of im in ToTensor() to guarantee array input
            # to torch.from_numpy is writable. Using im is a suspected cause of
            # https://github.com/pixeltable/pixeltable/issues/69
            return torchvision.transforms.ToTensor()(arr)
        elif self.column_types[k].is_json_type():
            assert isinstance(v, str)
            return json.loads(v)
        elif self.column_types[k].is_array_type():
            assert isinstance(v, np.ndarray)
            if not v.flags['WRITEABLE']:
                v = v.copy()
            assert v.flags['WRITEABLE']
            return v
        elif self.column_types[k].is_date_type():
            # pytorch default collation only supports numeric types
            assert isinstance(v, datetime.date)
            return v.toordinal()
        elif self.column_types[k].is_timestamp_type():
            # pytorch default collation only supports numeric types
            assert isinstance(v, datetime.datetime)
            return v.timestamp()
        else:
            assert not isinstance(v, np.ndarray)  # all array outputs should be handled above
            return v

    def __iter__(self) -> Iterator[dict[str, Any]]:
        from pixeltable.utils import arrow

        worker_info = torch.utils.data.get_worker_info()

        part_list: Sequence[int] = range(len(self.part_metadata))
        if worker_info is not None:
            part_list = [i for i in part_list if (i % worker_info.num_workers) == worker_info.id]

        for part_no in part_list:
            pqf = parquet.ParquetFile(self.part_metadata[part_no])
            for batch in pqf.iter_batches():
                for tup in arrow.iter_tuples(batch):
                    yield {k: self._unmarshall(k, v) for k, v in tup.items()}



from typing import Optional

import sqlalchemy as sql

from pixeltable import type_system as ts

from .expr import DataRow, Expr
from .literal import Literal
from .row_builder import RowBuilder
from .sql_element_cache import SqlElementCache


class TypeCast(Expr):
    """
    An `Expr` that represents a type conversion from an underlying `Expr` to
    a specified `ColumnType`.
    """

    def __init__(self, underlying: Expr, new_type: ts.ColumnType):
        super().__init__(new_type)
        self.components: list[Expr] = [underlying]
        self.id: Optional[int] = self._create_id()

    def _equals(self, other: 'TypeCast') -> bool:
        # `TypeCast` has no properties beyond those captured by `Expr`.
        return True

    @property
    def _op1(self) -> Expr:
        return self.components[0]

    def sql_expr(self, _: SqlElementCache) -> Optional[sql.ColumnElement]:
        """
        sql_expr() is unimplemented for now, in order to sidestep potentially thorny
        questions about consistency of doing type conversions in both Python and Postgres.
        """
        return None

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        original_val = data_row[self._op1.slot_idx]
        data_row[self.slot_idx] = self.col_type.create_literal(original_val)

    def as_literal(self) -> Optional[Literal]:
        op1_lit = self._op1.as_literal()
        if op1_lit is None:
            return None
        if not (
            self.col_type.is_numeric_type() and (op1_lit.col_type.is_numeric_type() or op1_lit.col_type.is_bool_type())
        ):
            return None

        op1_val = op1_lit.val
        if self.col_type.is_int_type():
            return Literal(int(op1_val), self.col_type)
        elif self.col_type.is_float_type():
            return Literal(float(op1_val), self.col_type)
        return None

    def _as_dict(self) -> dict:
        return {'new_type': self.col_type.as_dict(), **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: dict, components: list[Expr]) -> 'TypeCast':
        assert 'new_type' in d
        assert len(components) == 1
        return cls(components[0], ts.ColumnType.from_dict(d['new_type']))

    def __repr__(self) -> str:
        return f'{self._op1}.astype({self.col_type._to_str(as_schema=True)})'



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs)
that wrap various endpoints from the Groq API. In order to use them, you must
first `pip install groq` and configure your Groq credentials, as described in
the [Working with Groq](https://pixeltable.readme.io/docs/working-with-groq) tutorial.
"""

from typing import TYPE_CHECKING, Any, Optional

import pixeltable as pxt
from pixeltable import exprs
from pixeltable.env import Env, register_client
from pixeltable.utils.code import local_public_names

from .openai import _openai_response_to_pxt_tool_calls

if TYPE_CHECKING:
    import groq


@register_client('groq')
def _(api_key: str) -> 'groq.AsyncGroq':
    import groq

    return groq.AsyncGroq(api_key=api_key)


def _groq_client() -> 'groq.AsyncGroq':
    return Env.get().get_client('groq')


@pxt.udf(resource_pool='request-rate:groq')
async def chat_completions(
    messages: list[dict[str, str]],
    *,
    model: str,
    model_kwargs: Optional[dict[str, Any]] = None,
    tools: Optional[list[dict[str, Any]]] = None,
    tool_choice: Optional[dict[str, Any]] = None,
) -> dict:
    """
    Chat Completion API.

    Equivalent to the Groq `chat/completions` API endpoint.
    For additional details, see: <https://console.groq.com/docs/api-reference#chat-create>

    Request throttling:
    Applies the rate limit set in the config (section `groq`, key `rate_limit`). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install groq`

    Args:
        messages: A list of messages comprising the conversation so far.
        model: ID of the model to use. (See overview here: <https://console.groq.com/docs/models>)
        model_kwargs: Additional keyword args for the Groq `chat/completions` API.
            For details on the available parameters, see: <https://console.groq.com/docs/api-reference#chat-create>

    Returns:
        A dictionary containing the response and other metadata.

    Examples:
        Add a computed column that applies the model `llama3-8b-8192`
        to an existing Pixeltable column `tbl.prompt` of the table `tbl`:

        >>> messages = [{'role': 'user', 'content': tbl.prompt}]
        ... tbl.add_computed_column(response=chat_completions(messages, model='llama3-8b-8192'))
    """
    if model_kwargs is None:
        model_kwargs = {}

    Env.get().require_package('groq')

    if tools is not None:
        model_kwargs['tools'] = [{'type': 'function', 'function': tool} for tool in tools]

    if tool_choice is not None:
        if tool_choice['auto']:
            model_kwargs['tool_choice'] = 'auto'
        elif tool_choice['required']:
            model_kwargs['tool_choice'] = 'required'
        else:
            assert tool_choice['tool'] is not None
            model_kwargs['tool_choice'] = {'type': 'function', 'function': {'name': tool_choice['tool']}}

    if tool_choice is not None and not tool_choice['parallel_tool_calls']:
        model_kwargs['parallel_tool_calls'] = False

    result = await _groq_client().chat.completions.create(
        messages=messages,  # type: ignore[arg-type]
        model=model,
        **model_kwargs,
    )
    return result.model_dump()


def invoke_tools(tools: pxt.func.Tools, response: exprs.Expr) -> exprs.InlineDict:
    """Converts an OpenAI response dict to Pixeltable tool invocation format and calls `tools._invoke()`."""
    return tools._invoke(_openai_response_to_pxt_tool_calls(response))


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



from __future__ import annotations

from typing import Generic, Iterable, Iterator, Optional, TypeVar

from .expr import Expr

T = TypeVar('T', bound='Expr')


class ExprSet(Generic[T]):
    """
    A set that also supports indexed lookup (by slot_idx and Expr.id). Exprs are uniquely identified by Expr.id.
    """

    exprs: dict[int, T]  # key: Expr.id
    exprs_by_idx: dict[int, T]  # key: slot_idx

    def __init__(self, elements: Optional[Iterable[T]] = None):
        self.exprs = {}
        self.exprs_by_idx = {}
        if elements is not None:
            for e in elements:
                self.add(e)

    def add(self, expr: T) -> None:
        if expr.id in self.exprs:
            return
        self.exprs[expr.id] = expr
        if expr.slot_idx is None:
            return
        self.exprs_by_idx[expr.slot_idx] = expr

    def update(self, *others: Iterable[T]) -> None:
        for other in others:
            for e in other:
                self.add(e)

    def __contains__(self, item: T) -> bool:
        return item.id in self.exprs

    def __len__(self) -> int:
        return len(self.exprs)

    def __iter__(self) -> Iterator[T]:
        return iter(self.exprs.values())

    def __getitem__(self, index: object) -> Optional[T]:
        """Indexed lookup by slot_idx or Expr.id."""
        assert isinstance(index, (int, Expr))
        if isinstance(index, int):
            # return expr with matching slot_idx
            return self.exprs_by_idx.get(index)
        else:
            return self.exprs.get(index.id)

    def issuperset(self, other: ExprSet[T]) -> bool:
        return self.exprs.keys() >= other.exprs.keys()

    def __ge__(self, other: ExprSet[T]) -> bool:
        return self.issuperset(other)

    def __le__(self, other: ExprSet[T]) -> bool:
        return other.issuperset(self)

    def union(self, *others: Iterable[T]) -> ExprSet[T]:
        result = ExprSet(self.exprs.values())
        result.update(*others)
        return result

    def __or__(self, other: ExprSet[T]) -> ExprSet[T]:
        return self.union(other)

    def difference(self, *others: Iterable[T]) -> ExprSet[T]:
        id_diff = set(self.exprs.keys()).difference(e.id for other_set in others for e in other_set)
        return ExprSet(self.exprs[id] for id in id_diff)

    def __sub__(self, other: ExprSet[T]) -> ExprSet[T]:
        return self.difference(other)

    def __add__(self, other: ExprSet) -> ExprSet:
        exprs = self.exprs.copy()
        exprs.update(other.exprs)
        return ExprSet(exprs.values())
